import os
import pathlib
import warnings
from collections.abc import Callable

import numpy as np
import numpy.typing as npt

from .. import colormap
from .. import debug as debug
from .. import functions as fn
from .. import functions_qimage
from .. import getConfigOption
from ..Point import Point
from ..Qt import QtCore, QtGui, QtWidgets
from ..util.cupy_helper import getCupy
from .GraphicsObject import GraphicsObject

translate = QtCore.QCoreApplication.translate

__all__ = ['ImageItem']


class ImageItem(GraphicsObject):
    """
    Graphics object used to display image data.

    ImageItem can render images with 1, 3 or 4 channels, use lookup tables to apply
    false colors to images, and users can either set levels limits, or rely on
    the auto-sampling.

    Performance can vary wildly based on the attributes of the inputs provided, see
    :ref:`performance <ImageItem_performance>` for guidance if performance is an 
    important factor.

    There is optional `numba` and `cupy` support.

    **Bases:** :class:`pyqtgraph.GraphicsObject`

    Parameters
    ----------
    image : np.ndarray or None, default None
        Image data.
    **kargs : dict, optional
        Arguments directed to `setImage` and `setOpts`, refer to each method for
        documentation for possible arguments.

    Signals
    -------
    sigImageChanged: :class:`Signal`
        Emitted when the image is changed.
    sigRemoveRequested: :class:`Signal`
        Emitted when there is a request to remove the image. Signal emits the instance
        of the :class:`~pyqtgraph.ImageItem` whose removal is requested.

    See Also
    --------
    setImage :
        For descriptions of available keyword arguments.
    setOpts :
        For information on supported formats.
    """
    sigImageChanged = QtCore.Signal()
    sigRemoveRequested = QtCore.Signal(object) 

    def __init__(self, image: np.ndarray | None=None, **kargs):
        super().__init__()
        self.menu = None
        self.image = None   ## original image data
        self.qimage = None  ## rendered image for display

        self.paintMode = None
        self.levels = None  ## [min, max] or [[redMin, redMax], ...]
        self.lut = None
        self.autoDownsample = False
        self._nanPolicy = 'propagate'
        self._colorMap = None # This is only set if a color map is assigned directly
        self._lastDownsample = (1, 1)
        self._processingBuffer = None
        self._displayBuffer = None
        self._renderRequired = True
        self._unrenderable = False
        self._xp = None  # either numpy or cupy, to match the image data
        self._defferedLevels = None
        self._imageHasNans = None    # None : not yet known
        self._imageNanLocations = None
        self._defaultAutoLevels = True

        self.axisOrder = getConfigOption('imageAxisOrder')
        self._dataTransform = self._inverseDataTransform = QtGui.QTransform()
        self._update_data_transforms( self.axisOrder ) # install initial transforms

        self.drawKernel = None
        self.border = None
        self.removable = False

        if image is not None:
            self.setImage(image, **kargs)
        else:
            self.setOpts(**kargs)

    def setCompositionMode(self, mode: QtGui.QPainter.CompositionMode):
        """
        Change the composition mode of the item, useful when overlaying multiple items.
        
        Parameters
        ----------
        mode : :class:`QPainter.CompositionMode <QPainter.CompositionMode>`
            Composition of the item, often used when overlaying items.  Common
            options include:

            * `QPainter.CompositionMode.CompositionMode_SourceOver`
              Image replaces the background if it is opaque. Otherwise, it uses the
              alpha channel to blend the image with the background, default.
            * `QPainter.CompositionMode.CompositionMode_Overlay` Image color is
              mixed with the background color to reflect the lightness or darkness of
              the background.
            * `QPainter.CompositionMode.CompositionMode_Plus` Both the alpha and
              color of the image and background pixels are added together.
            * `QPainter.CompositionMode.CompositionMode_Plus` The output is the
              image color multiplied by the background.

            See :class:`QPainter.CompositionMode <QPainter.CompositionMode>` in the Qt
            documentation for more options and details.
        
        See Also
        --------
        :class:`QPainter.CompositionMode <QPainter.CompositionMode>` :
            Details all the possible composition mode options accepted.
        """
        self.paintMode = mode
        self.update()

    def setBorder(self, b):
        """
        Define the color of the border drawn around the image.

        Parameters
        ----------
        b : color_like
            Accepts all arguments supported by :func:`~pyqtgraph.mkPen`.
        """
        self.border = fn.mkPen(b)
        self.update()

    def width(self) -> int | None:
        if self.image is None:
            return None
        axis = 0 if self.axisOrder == 'col-major' else 1
        return self.image.shape[axis]

    def height(self) -> int | None:
        if self.image is None:
            return None
        axis = 1 if self.axisOrder == 'col-major' else 0
        return self.image.shape[axis]

    def channels(self) -> int | None:
        if self.image is None:
            return None
        return self.image.shape[2] if self.image.ndim == 3 else 1

    def boundingRect(self) -> QtCore.QRectF:
        if self.image is None:
            return QtCore.QRectF(0., 0., 0., 0.)
        if (width := self.width()) is None:
            width = 0.
        if (height := self.height()) is None:
            height = 0.
        return QtCore.QRectF(0., 0., float(width), float(height))

    def setAutoLevels(self, bState: bool):
        """
        Controls whether automatic image scaling takes place for this ImageItem,
        if not otherwise overridden by ``autoLevels`` or ``levels`` keyword
        arguments in a call to :func:`~pyqtgraph.ImageItem.setImage`.
        """
        self._defaultAutoLevels = bState

    def setLevels(self, levels: npt.ArrayLike | None, update: bool=True):
        """
        Set image scaling levels.

        Calling this method, even with ``levels=None`` will disable auto leveling 
        which is equivalent to :meth:`setImage` with ``autoLevels=False``.
        
        Parameters
        ----------
        levels : array_like or None
            Sets the numerical values that correspond to the limits of the color range.

            * ``[blackLevel, whiteLevel]`` 
                sets black and white levels for monochrome data and can be used with a
                lookup table.
            * ``[[minR, maxR], [minG, maxG], [minB, maxB]]``
                sets individual scaling for RGB values. Not compatible with lookup 
                tables.
            * ``None``
                Disables the application of levels, but setting to ``None`` prevents
                the auto-levels mechanism from sampling the image.  Not compatible with
                images that use floating point dtypes.
        update : bool, default True
            Update the image immediately to reflect the new levels.

        See Also
        --------
        pyqtgraph.functions.makeARGB
            For more details on how levels are applied.
        """
        if self._xp is None:
            self.levels = levels
            self._defferedLevels = levels
            return
        if levels is not None:
            levels = self._xp.asarray(levels)
        self.levels = levels
        if update:
            self.updateImage()

    def getLevels(self) -> np.ndarray | None:
        """
        Return the array representing the current level settings.

        See :meth:`setLevels`. When `autoLevels` is active, the format is
        ``[blackLevel, whiteLevel]``.

        Returns
        -------
        np.ndarray or None
            The value that the levels are set to.
        """
        return self.levels
        
    def setColorMap(self, colorMap: colormap.ColorMap | str):
        """
        Set a color map for false color display of a monochrome image.

        Parameters
        ----------
        colorMap : :class:`~pyqtgraph.ColorMap` or `str`
            A string argument will be passed to
            :func:`colormap.get() <pyqtgraph.colormap.get>`.
        
        Raises
        ------
        TypeError
            Raised when `colorMap` is not of type `str` or :class:`~pyqtgraph.ColorMap`.
        """
        if isinstance(colorMap, colormap.ColorMap):
            self._colorMap = colorMap
        elif isinstance(colorMap, str):
            self._colorMap = colormap.get(colorMap)
        else:
            raise TypeError("'colorMap' argument must be ColorMap or string")
        self.setLookupTable( self._colorMap.getLookupTable(nPts=256) )
        
    def getColorMap(self) -> colormap.ColorMap | None:
        """
        Retrieve the :class:`~pyqtgraph.ColorMap` object currently used.

        Returns
        -------
        ColorMap or None 
            The assigned :class:`~pyqtgraph.ColorMap`, or `None` if not available.
        """
        return self._colorMap
        
    def setLookupTable(self, lut: npt.ArrayLike | Callable, update: bool=True):
        """
        Set lookup table `lut` to use for false color display of a monochrome image.

        Ordinarily, this table is supplied by a :class:`~pyqtgraph.HistogramLUTItem`,
        :class:`~pyqtgraph.GradientEditorItem` or :class:`~pyqtgraph.ColorBarItem`.

        Parameters
        ----------
        lut : array_like or callable
            If `lut` is an np.ndarray, ensure the dtype is `np.uint8`. Alternatively
            can be a callable that accepts the current image as an argument and
            returns the lookup table to use. Support for callable will be removed
            in a future version of pyqtgraph.
        update : bool, default True
            Update the intermediate image.
    
        See Also
        --------
        :func:`pyqtgraph.functions.makeARGB`
            See this function for more information on how this is used.
        :meth:`ColorMap.getLookupTable <pyqtgraph.ColorMap.getLookupTable>`
            Can construct a lookup table from a :class:`~pyqtgraph.ColorMap` object.

        Notes
        -----
        For performance reasons, if not passing a callable, every effort should be made
        to keep the number of entries to `<= 256`.
        """

        if lut is not self.lut:
            if self._xp is not None:
                lut = self._ensure_proper_substrate(lut, self._xp)
            self.lut = lut
            if update:
                self.updateImage()

    @staticmethod
    def _ensure_proper_substrate(data: Callable | npt.ArrayLike, substrate) -> np.ndarray:
        if data is None or isinstance(data, (Callable, substrate.ndarray)):
            return data
        cupy = getCupy()
        if substrate == cupy and not isinstance(data, cupy.ndarray):
            data = cupy.asarray(data)
        elif substrate == np:
            if cupy is not None and isinstance(data, cupy.ndarray):
                data = data.get()
            else:
                data = np.asarray(data)
        return data

    def setAutoDownsample(self, active: bool=True):
        """
        Control automatic downsampling for this ImageItem.

        Parameters
        ----------
        active : bool, default True
            If `active` is ``True``, the image is automatically downsampled to match
            the screen resolution. This improves performance for large images and
            reduces aliasing. If `autoDownsample` is not specified, then ImageItem will
            choose whether to downsample the image based on its size. ``False``
            disables automatic downsampling.
        """
        self.autoDownsample = active
        self._renderRequired = True
        self.update()

    def nanPolicy(self) -> str:
        """
        Retrieve the string representing the current NaN policy.

        See :meth:setNanPolicy.

        Returns
        -------
        { 'propagate', 'omit' }
            The NaN policy that this ImageItem uses during downsampling.
        """
        return self._nanPolicy

    def setNanPolicy(self, nanPolicy: str):
        """
        Control how NaN values are handled during downsampling for this ImageItem.

        Parameters
        ----------
        nanPolicy : { 'propagate', 'omit' }
            If 'nanPolicy' is 'ignore', NaNs are automatically ignored during
            downsampling, at the expense of performance. If 'nanPolicy' is 'propagate',
            NaNs are kept during downsampling. Unless a different policy was specified,
            a new ImageItem is created with ``nanPolicy='propagate'``.
        """
        if nanPolicy not in ['propagate', 'omit']:
            raise ValueError(f"{nanPolicy=} must be one of {'propagate', 'omit'}")
        self._nanPolicy = nanPolicy
        self._renderRequired = True
        self.update()

    def setOpts(self, update: bool=True, **kwargs):
        """
        Set display and processing options for this ImageItem.

        :class:`~pyqtgraph.ImageItem` and :meth:`setImage` support all keyword
        arguments listed here.

        Parameters
        ----------
        update : bool, default True
            Controls if image immediately updates to reflect the new options.

        **kwargs : dict, optional
            Extra arguments that are directed to the respective methods.  Expected
            keys include:

            * `autoDownsample` whose value is directed to :meth:`setAutoDownsample`
            * `nanPolicy` whose value is directed to :meth:`setNanPolicy`
            * `axisOrder`, which needs to be one of {'row-major', 'col-major'},
              determines the relationship between the numpy axis and visual axis
              of the data.
            * `border`, whose value is directed to :meth:`setBorder`
            * `colorMap`, whose value is directed to :meth:`setColorMap`
            * `compositionMode`, whose value is directed to :meth:`setCompositionMode`
            * `levels` whose value is directed to :meth:`setLevels`
            * `lut`, whose value  is directed to :meth:`setLookupTable`
            * `opacify` whose value is directed to
              :meth:`QGraphicsItem.setOpacity <QGraphicsItem.setOpacity>`
            * `rect` whose value is directed to :meth:`setRect`
            * `removable` boolean, determines if the context menu is available
        
        See Also
        --------
        :meth:`setAutoDownsample`
            Accepts the value of ``kwargs['autoDownsample']``.
        :meth:`setAutoLevels`
            Accepts the value of ``kwargs['autoLevels']``.
        :meth:`setNanPolicy`
            Accepts the value of ``kwargs['nanPolicy']``.
        :meth:`setBorder`
            Accepts the value of ``kwargs['border']``.
        :meth:`setColorMap`
            Accepts the value of ``kwargs['colorMap']``.
        :meth:`setCompositionMode`
            Accepts the value of ``kwargs['compositionMode']``.
        :meth:`setImage`
            Accepts the value of ``kwargs['image']``.
        :meth:`setLevels`
            Accepts the value of ``kwargs['levels']``.
        :meth:`setLookupTable`
            Accepts the value of ``kwargs['lut']``.
        :meth:`QGraphicsItem.setOpacity <QGraphicsItem.setOpacity>`
            Accepts the value of ``kwargs['opacity']``.
        :meth:`setRect`
            Accepts the value of ``kwargs['rect']``.
        """
        if 'axisOrder' in kwargs:
            val = kwargs['axisOrder']            
            if val not in ('row-major', 'col-major'):
                raise ValueError("axisOrder must be either 'row-major' or 'col-major'")
            self.axisOrder = val
            self._update_data_transforms(self.axisOrder) # update cached transforms
        if 'colorMap' in kwargs:
            self.setColorMap(kwargs['colorMap'])
        if 'lut' in kwargs:
            self.setLookupTable(kwargs['lut'], update=update)
        if 'levels' in kwargs:
            self.setLevels(kwargs['levels'], update=update)
        #if 'clipLevel' in kargs:
            #self.setClipLevel(kargs['clipLevel'])
        if 'opacity' in kwargs:
            self.setOpacity(kwargs['opacity'])
        if 'compositionMode' in kwargs:
            self.setCompositionMode(kwargs['compositionMode'])
        if 'border' in kwargs:
            self.setBorder(kwargs['border'])
        if 'removable' in kwargs:
            self.removable = kwargs['removable']
            self.menu = None
        if 'autoDownsample' in kwargs:
            self.setAutoDownsample(kwargs['autoDownsample'])
        if 'autoLevels' in kwargs:
            self.setAutoLevels(kwargs['autoLevels'])
        if 'nanPolicy' in kwargs:
            self.setNanPolicy(kwargs['nanPolicy'])
        if 'rect' in kwargs:
            self.setRect(kwargs['rect'])
        if update:
            self.update()

    def setRect(self, *args):
        """
        Set view rectangle for the :class:`~pyqtgraph.ImageItem` to occupy.

        In addition to accepting a :class:`QRectF`, you can pass the numerical values
        representing the  `x, y, w, h`, where `x, y` represent the x, y coordinates
        of the top left corner, and `w` and `h` represent the width and height
        respectively.

        Parameters
        ----------
        *args : tuple
            Contains one of :class:`QRectF`, :class:`QRect`, or arguments that can be
            used to construct :class:`QRectF`.

        See Also
        --------
        :class:`QRectF` :
            See constructor methods for allowable `*args`.

        Notes
        -----
        This method cannot be used before an image is assigned. See the
        :ref:`examples <ImageItem_examples>` for how to manually set transformations.
        """
        if not args:
            # reset scaling and rotation when called without argument
            self.resetTransform()
            return
        if isinstance(args[0], (QtCore.QRectF, QtCore.QRect)):
            rect = args[0] # use QRectF or QRect directly
        else:
            if hasattr(args[0],'__len__'):
                args = args[0] # promote tuple or list of values
            # QRectF(x,y,w,h), but also accepts other initializers
            rect = QtCore.QRectF( *args ) 
        tr = QtGui.QTransform()
        tr.translate(rect.left(), rect.top())

        if (width := self.width()) is None:
            width = 1.

        if (height := self.height()) is None:
            height = 1.

        tr.scale(rect.width() / width, rect.height() / height)
        self.setTransform(tr)

    def clear(self):
        """
        Clear the assigned image.
        """
        self.image = None
        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.update()

    def _buildQImageBuffer(self, shape: tuple[int, int, int]):
        self._displayBuffer = np.empty(shape[:2] + (4,), dtype=np.ubyte)
        if self._xp == getCupy():
            self._processingBuffer = self._xp.empty(
                shape[:2] + (4,),
                dtype=self._xp.ubyte
            )
        else:
            self._processingBuffer = self._displayBuffer
        self.qimage = None

    def setImage(
            self,
            image: np.ndarray | None=None,
            autoLevels: bool | None=None,
            levelSamples: int = 65536,
            **kwargs
        ):
        """
        Update the image displayed by this ImageItem.
        
        All keywords supported by :meth:`setOpts` are also allowed here.

        Parameters
        ----------
        image : np.ndarray or None, default None
            Image data given as NumPy array with an integer or floating point dtype of
            any bit depth. A 2-dimensional array describes single-valued
            (monochromatic) data. A 3-dimensional array is used to give individual
            color components. The third dimension must be of length 3 (RGB) or 4
            (RGBA). ``np.nan`` values are treated as transparent pixels.
        autoLevels : bool or None, default None
            If ``True``, ImageItem will automatically select levels based on the maximum
            and minimum values encountered in the data. For performance reasons, this
            search sub-samples the images and may miss individual bright or dark points
            in the data set. If ``False``, the search will be omitted. If ``None``, the
            value set by :func:`~pyqtgraph.ImageItem.setOpts` is used, unless a ``levels``
            keyword argument is given, which implies `False`.
        levelSamples : int, default 65536
            Only used when ``autoLevels is None``.  When determining minimum and
            maximum values, ImageItem only inspects a subset of pixels no larger than
            this number. Setting this larger than the total number of pixels considers
            all values. See `quickMinMax`.
        **kwargs : dict, optional
            Extra arguments that are passed to `setOpts`.

        See Also
        --------
        quickMinMax
            See this method for how levelSamples value is utilized.
        :func:`pyqtgraph.functions.makeARGB`
            See this function for how image data is modified prior to rendering.

        Notes
        -----
        For backward compatibility, image data is assumed to be in column-major order
        (column, row) by default. However, most data is stored in row-major order
        (row, column). It can either be transposed before assignment
        
        .. code-block:: python

            imageitem.setImage(imagedata.T)
        
        or the interpretation of the data can be changed locally through the
        `axisOrder` keyword or by changing the `imageAxisOrder`
        :ref:`global configuration option <apiref_config>`.
        """
        profile = debug.Profiler()

        gotNewData = False
        if image is None:
            if self.image is None:
                return
        else:
            old_xp = self._xp
            cp = getCupy()
            self._xp = cp.get_array_module(image) if cp else np
            gotNewData = True
            processingSubstrateChanged = old_xp != self._xp
            if processingSubstrateChanged:
                self._processingBuffer = None
            shapeChanged = (
                processingSubstrateChanged or
                self.image is None or
                image.shape != self.image.shape
            )
            image = image.view()
            self.image = image
            self._imageHasNans = None
            self._imageNanLocations = None
            if 'autoDownsample' not in kwargs and (
                self.image.shape[0] > 2**15-1 or self.image.shape[1] > 2**15-1
            ):
                kwargs['autoDownsample'] = True
            if shapeChanged:
                self.prepareGeometryChange()
                self.informViewBoundsChanged()

        profile()

        if autoLevels is None:
            autoLevels = False if 'levels' in kwargs else self._defaultAutoLevels

        if autoLevels:
            mn, mx = self.quickMinMax( targetSize=levelSamples )
            # mn and mx can still be NaN if the data is all-NaN
            if mn == mx or self._xp.isnan(mn) or self._xp.isnan(mx):
                mn = 0
                mx = 255
            kwargs['levels'] = self._xp.asarray((mn,mx))

        profile()

        self.setOpts(update=False, **kwargs)

        profile()

        self._renderRequired = True
        self.update()

        profile()

        if gotNewData:
            self.sigImageChanged.emit()
        if self._defferedLevels is not None:
            levels = self._defferedLevels
            self._defferedLevels = None
            self.setLevels((levels))

    def _update_data_transforms(self, axisOrder: str='col-major'):
        """
        Set up the transforms needed to map between input array and display.

        Parameters
        ----------
        axisOrder : { 'col-major', 'row-major' }
            The axis order to update the data transformation to.
        """
        self._dataTransform = QtGui.QTransform()
        self._inverseDataTransform = QtGui.QTransform()
        if self.axisOrder == 'row-major': # transpose both
            self._dataTransform.scale(1, -1)
            self._dataTransform.rotate(-90)
            self._inverseDataTransform.scale(1, -1)
            self._inverseDataTransform.rotate(-90)

    def dataTransform(self):
        """
        Get the transform mapping image array to local coordinate system.

        This transform corrects for the transposition that occurs when image data is
        interpreted in row-major order.
        
        :meta private:

        Returns
        -------
        :class:`QTransform`
            The transform that is used for mapping.
        """
        # Might eventually need to account for downsampling / clipping here
        # transforms are updated in setOpts call.
        return self._dataTransform

    def inverseDataTransform(self) -> QtGui.QTransform:
        """
        Get the transform mapping local coordinate system to image array.

        :meta private:

        Returns
        -------
        :class:`QTransform`
            The transform that is used for mapping.
        
        See Also
        --------
        dataTransform
            See dataTransform() for more information. 
        """
        # transforms are updated in setOpts call.
        return self._inverseDataTransform

    def mapToData(self, obj):
        return self._inverseDataTransform.map(obj)

    def mapFromData(self, obj):
        return self._dataTransform.map(obj)

    def quickMinMax(self, targetSize: int=1_000_000) -> tuple[float, float]:
        """
        Estimate the min and max values of the image data by sub-sampling.

        Sampling is performed at regular strides chosen to evaluate a number of 
        samples equal to or less than `targetSize`.  Returns the estimated min and max
        values of the image data.

        Parameters
        ----------
        targetSize : int, default 1_000_000
            The number of pixels to downsample the image to.

        Returns
        -------
        float, float
            Estimated minimum and maximum values of the image data.
        """

        data = self.image
        if data is None:
            # image hasn't been set yet
            return 0., 0.
        targetSize = max(targetSize, 2) # keep at least 2 pixels
        while True:
            h, w = data.shape[:2]
            if h * w <= targetSize: break
            data = data[::2, ::] if h > w else data[::, ::2]
        return self._xp.nanmin(data), self._xp.nanmax(data)

    def updateImage(self, *args, **kargs):
        defaults = {
            'autoLevels': False,
        } | kargs
        return self.setImage(*args, **defaults)

    def render(self):
        # Convert data to QImage for display.
        self._unrenderable = True
        if self.image is None or self.image.size == 0:
            return

        # Request a lookup table if this image has only one channel
        if self.image.ndim == 2 or self.image.shape[2] == 1:
            self.lut = self._ensure_proper_substrate(self.lut, self._xp)
            if isinstance(self.lut, Callable):
                lut = self._ensure_proper_substrate(
                    self.lut(self.image, 256),
                    self._xp
                )
            else:
                lut = self.lut
        else:
            lut = None

        if self._imageHasNans is None:
            # awkward, but fastest numpy native nan evaluation
            self._imageHasNans = (
                self.image.dtype.kind == 'f' and
                self._xp.isnan(self.image.min())
            )
            self._imageNanLocations = None

        image = self.image

        if self.autoDownsample:
            xds, yds = self._computeDownsampleFactors()
            if xds is None:
                return

            axes = [1, 0] if self.axisOrder == 'row-major' else [0, 1]
            nan_policy = self._nanPolicy if self._imageHasNans else 'propagate'
            image = fn.downsample(image, xds, axis=axes[0], nanPolicy=nan_policy)
            image = fn.downsample(image, yds, axis=axes[1], nanPolicy=nan_policy)
            self._lastDownsample = (xds, yds)

            # changes in view transform cause changes in downsampling factors,
            # which invalidates any previously calculated nan locations
            self._imageNanLocations = None

            # Check if downsampling reduced the image size to zero due to inf values.
            if image.size == 0:
                return

        # Convert single-channel image to 2D array
        if image.ndim == 3 and image.shape[-1] == 1:
            image = image[..., 0]

        # Assume images are in column-major order for backward compatibility
        # (most images are in row-major order)
        if self.axisOrder == 'col-major':
            image = image.swapaxes(0, 1)

        levels = self.levels


        qimage = None

        if lut is not None and lut.dtype != self._xp.uint8:
            # try_make_image() assumes that lut is of type uint8.
            # It is considered a usage error if that is not the case.
            # However, the makeARGB() code-path has previously allowed such
            # a usage to work. Rather than fail outright, we delegate this
            # case to makeARGB().
            warnings.warn(
                ("Using non-uint8 LUTs is an undocumented accidental feature and may "
                "be removed at some point in the future. Please open an issue if you "
                "instead believe this to be worthy of protected inclusion in "
                "pyqtgraph."),
                DeprecationWarning,
                stacklevel=2
            )

        elif not self._imageHasNans:
            qimage = functions_qimage.try_make_qimage(image, levels=levels, lut=lut)

        elif image.ndim in (2, 3):
            # float images with nans
            if self._imageNanLocations is None:
                # the number of nans is expected to be small
                nanmask = self._xp.isnan(image)
                if nanmask.ndim == 3:
                    nanmask = nanmask.any(axis=2)
                self._imageNanLocations = nanmask.nonzero()
            qimage = functions_qimage.try_make_qimage(
                image,
                levels=levels,
                lut=lut,
                transparentLocations=self._imageNanLocations
            )

        if qimage is not None:
            self._processingBuffer = None
            self._displayBuffer = None
            self.qimage = qimage
            self._renderRequired = False
            self._unrenderable = False
            return

        if (
            self._processingBuffer is None or 
            self._processingBuffer.shape[:2] != image.shape[:2]
        ):
            self._buildQImageBuffer(image.shape)

        fn.makeARGB(image, lut=lut, levels=levels, output=self._processingBuffer)
        if self._xp == getCupy():
            self._processingBuffer.get(out=self._displayBuffer)
        self.qimage = fn.ndarray_to_qimage(
            self._displayBuffer,
            QtGui.QImage.Format.Format_ARGB32
        )

        self._renderRequired = False
        self._unrenderable = False

    def paint(self, painter, *args):
        profile = debug.Profiler()
        if self.image is None:
            return
        if self._renderRequired:
            self.render()
            if self._unrenderable:
                return
            profile('render QImage')
        if self.paintMode is not None:
            painter.setCompositionMode(self.paintMode)
            profile('set comp mode')

        shape = (
            self.image.shape[:2]
            if self.axisOrder == 'col-major'
            else self.image.shape[:2][::-1]
        )
        painter.drawImage(QtCore.QRectF(0,0,*shape), self.qimage)
        profile('p.drawImage')
        if self.border is not None:
            painter.setPen(self.border)
            painter.drawRect(self.boundingRect())

    def save(self, fileName: str | pathlib.Path, *args) -> None:
        """
        Save this image to file.

        Note that this saves the visible image, after scale/color changes, not the
        original data.

        Parameters
        ----------
        fileName : os.PathLike
            File path to save the image data to.
        *args : tuple
            Arguments that are passed to :meth:`QImage.save <QImage.save>`.
            
        See Also
        --------
        :meth:`QImage.save <QImage.save>` :
            ``*args`` is relayed to this method.
        """
        if self.qimage is None:
            return None

        if self._renderRequired:
            self.render()

        self.qimage.save(os.fsdecode(fileName), *args)

    def getHistogram(
        self,
        bins: str | int='auto',
        step: str | np.generic='auto',
        perChannel: bool=False,
        targetImageSize: int=200,
        **kwargs
    ) -> list[tuple[np.ndarray, np.ndarray]] | tuple[np.ndarray, np.ndarray] | tuple[None, None]:
        """
        Generate arrays containing the histogram values.

        Similar to :func:`numpy.histogram`

        Parameters
        ----------
        bins : int or str, default 'auto'
            The `bins` argument and any extra keyword arguments are passed to
            :func:`numpy.histogram()`. If ``bins == 'auto'``, a bin number is
            automatically chosen based on the image characteristics.
        step : int or str, default 'auto'
            The `step` argument causes pixels to be skipped when computing the
            histogram to save time. If `step` is 'auto', then a step is chosen such
            that the analyzed data has dimensions approximating `targetImageSize`
            for each axis.
        perChannel : bool, default False
            If ``True``, then a histogram is computed for each channel, and the output
            is a list of the results.
        targetImageSize : int, default 200
            This parameter is used if ``step == 'auto'``, If so, the `step` size is
            calculated by ``step = ceil(image.shape[0] / targetImageSize)``.
        **kwargs : dict, optional
            Dictionary of arguments passed to :func:`numpy.histogram()`.
        
        Returns
        -------
        numpy.ndarray, numpy.ndarray or None, None or list of tuple of numpy.ndarray, numpy.ndarray
            Returns `x` and `y` arrays containing the histogram values for the current
            image. For an explanation of the return format, see
            :func:`numpy.histogram()`.
            Returns ``[(numpy.ndarray, numpy.ndarray),...]`` if ``perChannel=True``, one
            element per channel.
            Returns ``(None, None)`` is there is no image, or image size is 0.

        Warns
        -----
        RuntimeWarning
            Emits when `targetHistogramSize` argument is passed in, which does nothing.

        See Also
        --------
        numpy.histogram :
            Describes return format in greater detail.
        numpy.histogram_bin_edges:
            Details the different string values accepted as the `bins` parameter.
        """        

        if 'targetHistogramSize' in kwargs:
            warnings.warn(
                "'targetHistogramSize' option is not used",
                RuntimeWarning,
                stacklevel=2
            )

        # This method is also used when automatically computing levels.
        if self.image is None or self.image.size == 0:
            return None, None
        if step == 'auto':
            step = (max(1, int(self._xp.ceil(self.image.shape[0] / targetImageSize))),
                    max(1, int(self._xp.ceil(self.image.shape[1] / targetImageSize))))
        if self._xp.isscalar(step):
            step = (step, step)
        stepData = self.image[::step[0], ::step[1]]

        if isinstance(bins, str) and bins == 'auto':
            mn = self._xp.nanmin(stepData).item()
            mx = self._xp.nanmax(stepData).item()
            if mx == mn:
                # degenerate image, arange will fail
                mx += 1
            if self._xp.isnan(mn) or self._xp.isnan(mx):
                # the data are all-nan
                return None, None
            if stepData.dtype.kind in "ui":
                # For integer data, we select the bins carefully to avoid aliasing
                step = int(self._xp.ceil((mx - mn) / 500.))
                bins = []
                if step > 0.0:
                    bins = self._xp.arange(mn, mx + 1.01 * step, step, dtype=int)
            else:
                # for float data, let numpy select the bins.
                bins = self._xp.linspace(mn, mx, 500)

            if len(bins) == 0:
                bins = self._xp.asarray((mn, mx))

        kwargs['bins'] = bins

        cp = getCupy()
        if perChannel:
            hist = []
            for i in range(stepData.shape[-1]):
                stepChan = stepData[..., i]
                stepChan = stepChan[self._xp.isfinite(stepChan)]
                h = self._xp.histogram(stepChan, **kwargs)
                if cp:
                    hist.append((cp.asnumpy(h[1][:-1]), cp.asnumpy(h[0])))
                else:
                    hist.append((h[1][:-1], h[0]))
            return hist
        else:
            stepData = stepData[self._xp.isfinite(stepData)]
            hist = self._xp.histogram(stepData, **kwargs)
            if cp:
                return cp.asnumpy(hist[1][:-1]), cp.asnumpy(hist[0])
            else:
                return hist[1][:-1], hist[0]

    def setPxMode(self, b: bool):
        """
        Set whether item ignores transformations and draws directly to screen pixels.

        Parameters
        ----------
        b : bool
            If ``True``, the item will not inherit any scale or rotation
            transformations from its parent items, but its position will be transformed
            as usual.

        See Also
        --------
        :class:`QGraphicsItem.GraphicsItemFlag <QGraphicsItem.GraphicsItemFlag>` :
            Read the description of `ItemIgnoresTransformations` for more information.
        """
        self.setFlag(
            QtWidgets.QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations,
            b
        )

    def setScaledMode(self):
        self.setPxMode(False)

    def getPixmap(self) -> QtGui.QPixmap | None:
        if self._renderRequired:
            self.render()
            if self._unrenderable:
                return None
        if self.qimage is None:
            return QtGui.QPixmap()
        return QtGui.QPixmap.fromImage(self.qimage)

    def pixelSize(self) -> tuple[float, float]:
        """
        Get the `x` and `y` size of each pixel in the view coordinate system.

        Returns
        -------
        float, float
            The `x` and `y` size of each pixel in scene space.
        """
        br = self.sceneBoundingRect()
        if self.image is None:
            return 1.,1.

        if (width := self.width()) is None:
            width = 0.
        if (height := self.height()) is None:
            height = 0.

        return br.width() / width, br.height() / height

    def viewTransformChanged(self):
        if self.autoDownsample:
            xds, yds = self._computeDownsampleFactors()
            if xds is None:
                self._renderRequired = True
                self._unrenderable = True
                return
            if (xds, yds) != self._lastDownsample:
                self._renderRequired = True
                self.update()

    def _computeDownsampleFactors(self) -> tuple[int, int]:
        # reduce dimensions of image based on screen resolution
        o = self.mapToDevice(QtCore.QPointF(0, 0))
        x = self.mapToDevice(QtCore.QPointF(1, 0))
        y = self.mapToDevice(QtCore.QPointF(0, 1))
        # scene may not be available yet
        if o is None:
            return 1, 1
        w = Point(x - o).length()
        h = Point(y - o).length()
        if w == 0 or h == 0:
            return 1, 1
        return max(1, int(1.0 / w)), max(1, int(1.0 / h))

    def mouseDragEvent(self, ev):
        if ev.button() != QtCore.Qt.MouseButton.LeftButton:
            ev.ignore()
            return
        elif self.drawKernel is not None:
            ev.accept()
            self.drawAt(ev.pos(), ev)

    def mouseClickEvent(self, ev):
        if (
            ev.button() == QtCore.Qt.MouseButton.RightButton and
            self.raiseContextMenu(ev)
        ):
            ev.accept()
        if self.drawKernel is not None and ev.button() == QtCore.Qt.MouseButton.LeftButton:
            self.drawAt(ev.pos(), ev)

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        if menu is None:
            return False
        if self.scene() is None:
            warnings.warn(
                (
                    "Attempting to raise a context menu with the GraphicsScene has "
                    "not been set. Returning None"
                ),
                RuntimeWarning,
                stacklevel=2
            )
            return None
        menu = self.scene().addParentContextMenus(self, menu, ev)
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(int(pos.x()), int(pos.y())))
        return True

    def getMenu(self):
        if self.menu is None:
            if not self.removable:
                return None
            self.menu = QtWidgets.QMenu()
            self.menu.setTitle(translate("ImageItem", "Image"))
            remAct = QtGui.QAction(translate("ImageItem", "Remove image"), self.menu)
            remAct.triggered.connect(self.removeClicked)
            self.menu.addAction(remAct)
            self.menu.remAct = remAct
        return self.menu

    def hoverEvent(self, ev):
        if not ev.isExit():
            if self.drawKernel is not None and ev.acceptDrags(
                QtCore.Qt.MouseButton.LeftButton
            ):
                # we don't use the click, but we also don't want anyone else to use it
                ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton)
                ev.acceptClicks(QtCore.Qt.MouseButton.RightButton)
            elif self.removable:
                # accept context menu clicks
                ev.acceptClicks(QtCore.Qt.MouseButton.RightButton)

    def tabletEvent(self, ev):
        pass

    def drawAt(self, pos, ev=None):
        if self.axisOrder == "col-major":
            pos = [int(pos.x()), int(pos.y())]
        else:
            pos = [int(pos.y()), int(pos.x())]
        dk = self.drawKernel
        kc = self.drawKernelCenter
        sx = [0,dk.shape[0]]
        sy = [0,dk.shape[1]]
        tx = [pos[0] - kc[0], pos[0] - kc[0]+ dk.shape[0]]
        ty = [pos[1] - kc[1], pos[1] - kc[1]+ dk.shape[1]]

        for i in [0,1]:
            dx1 = -min(0, tx[i])
            dx2 = min(0, self.image.shape[0]-tx[i])
            tx[i] += dx1+dx2
            sx[i] += dx1+dx2

            dy1 = -min(0, ty[i])
            dy2 = min(0, self.image.shape[1]-ty[i])
            ty[i] += dy1+dy2
            sy[i] += dy1+dy2

        ts = (slice(tx[0],tx[1]), slice(ty[0],ty[1]))
        ss = (slice(sx[0],sx[1]), slice(sy[0],sy[1]))
        mask = self.drawMask
        src = dk

        if isinstance(self.drawMode, Callable):
            self.drawMode(dk, self.image, mask, ss, ts, ev)
        else:
            src = src[ss]
            if self.drawMode == 'set':
                if mask is not None:
                    mask = mask[ss]
                    self.image[ts] = self.image[ts] * (1-mask) + src * mask
                else:
                    self.image[ts] = src
            elif self.drawMode == 'add':
                self.image[ts] += src
            else:
                raise ValueError(f"Unknown draw mode '{self.drawMode}'")
            self.updateImage()

    def setDrawKernel(self, kernel=None, mask=None, center=(0,0), mode='set'):
        self.drawKernel = kernel
        self.drawKernelCenter = center
        self.drawMode = mode
        self.drawMask = mask

    def removeClicked(self):
        ## Send remove event only after we have exited the menu event handler
        self.removeTimer = QtCore.QTimer()
        self.removeTimer.timeout.connect(self.emitRemoveRequested)
        self.removeTimer.start(0)

    def emitRemoveRequested(self):
        self.removeTimer.timeout.disconnect(self.emitRemoveRequested)
        self.sigRemoveRequested.emit(self)



import collections.abc
import os
import warnings
import weakref

from typing import Iterable

import numpy as np

from ... import functions as fn
from ... import icons
from ...Qt import QtCore, QtWidgets
from ...WidgetGroup import WidgetGroup
from ...widgets.FileDialog import FileDialog
from ..AxisItem import AxisItem
from ..ButtonItem import ButtonItem
from ..GraphicsWidget import GraphicsWidget
from ..InfiniteLine import InfiniteLine
from ..LabelItem import LabelItem
from ..LegendItem import LegendItem
from ..PlotCurveItem import PlotCurveItem
from ..PlotDataItem import PlotDataItem
from ..ViewBox import ViewBox

translate = QtCore.QCoreApplication.translate

from . import plotConfigTemplate_generic as ui_template

__all__ = ['PlotItem']


class PlotItem(GraphicsWidget):
    """
    GraphicsWidget implementing a standard 2D plotting area with axes.

    **Bases:** :class:`GraphicsWidget <pyqtgraph.GraphicsWidget>`
    
    This class provides the ViewBox-plus-axes that appear when using
    :func:`pg.plot() <pyqtgraph.plot>`, :class:`PlotWidget <pyqtgraph.PlotWidget>`,
    and :meth :`GraphicsLayout.addPlot() <pyqtgraph.GraphicsLayout.addPlot>`.

    It's main functionality is:

      - Manage placement of ViewBox, AxisItems, and LabelItems
      - Create and manage a list of PlotDataItems displayed inside the ViewBox
      - Implement a context menu with commonly used display and analysis options

    Use :func:`plot() <pyqtgraph.PlotItem.plot>` to create a new PlotDataItem and
    add it to the view. Use :func:`addItem() <pyqtgraph.PlotItem.addItem>` to
    add any QGraphicsItem to the view.
    
    This class wraps several methods from its internal ViewBox:
      - :func:`setXRange <pyqtgraph.ViewBox.setXRange>`
      - :func:`setYRange <pyqtgraph.ViewBox.setYRange>`
      - :func:`setRange <pyqtgraph.ViewBox.setRange>`
      - :func:`autoRange <pyqtgraph.ViewBox.autoRange>`
      - :func:`setDefaultPadding <pyqtgraph.ViewBox.setDefaultPadding>`
      - :func:`setXLink <pyqtgraph.ViewBox.setXLink>`
      - :func:`setYLink <pyqtgraph.ViewBox.setYLink>`
      - :func:`setAutoPan <pyqtgraph.ViewBox.setAutoPan>`
      - :func:`setAutoVisible <pyqtgraph.ViewBox.setAutoVisible>`
      - :func:`setLimits <pyqtgraph.ViewBox.setLimits>`
      - :func:`viewRect <pyqtgraph.ViewBox.viewRect>`
      - :func:`viewRange <pyqtgraph.ViewBox.viewRange>`
      - :func:`setMouseEnabled <pyqtgraph.ViewBox.setMouseEnabled>`
      - :func:`enableAutoRange <pyqtgraph.ViewBox.enableAutoRange>`
      - :func:`disableAutoRange <pyqtgraph.ViewBox.disableAutoRange>`
      - :func:`setAspectLocked <pyqtgraph.ViewBox.setAspectLocked>`
      - :func:`invertY <pyqtgraph.ViewBox.invertY>`
      - :func:`invertX <pyqtgraph.ViewBox.invertX>`
      - :func:`register <pyqtgraph.ViewBox.register>`
      - :func:`unregister <pyqtgraph.ViewBox.unregister>`
    
    The ViewBox itself can be accessed by calling :func:`getViewBox() <pyqtgraph.PlotItem.getViewBox>` 
    
    Parameters
    ----------
    parent : QObject or None, default None
        Parent :class:`QObject` assign to the :class:`~pyqtgraph.PlotItem`.
    name : str or None, default None
        Register the value for this view so that others may link to it.
    labels :  dict of str or None, default None
        A dictionary specifying the axis labels to display

        .. code-block:: python
                   
            {'left': (args), 'bottom': (args), ...}
        
        The name of each axis and the corresponding arguments are passed to 
        :meth:`PlotItem.setLabel() <pyqtgraph.PlotItem.setLabel>`
        Optionally, PlotItem my also be initialized with the keyword arguments left,
        right, top, or bottom to achieve the same effect.
    title : str
        Text to set the title of the PlotItem to.
    viewBox : :class:`~pyqtgraph.ViewBox` or None, default None
        Have the PlotItem use the provided :class:`~pyqtgraph.ViewBox`. If not
        specified, the PlotItem will be constructed with this as its ViewBox.
    axisItems : dict of {'left', 'bottom', 'right', 'top } and AxisItem or None
        Pass pre-constructed :class:`~pyqtgraph.AxisItem` objects to be used for the
        `left`, `bottom`, `right` or `top` axis.  Default is None.
    enableMenu : bool
        Toggle the enabling or disabling of the right-click context menu.
    **kwargs : dict, optional
        Any extra keyword arguments are passed to
        :func:`PlotItem.plot() <pyqtgraph.PlotItem.plot>`.
    
    Signals
    -------
    sigYRangeChanged : Signal
        Signal is emitted when the range on the y-axis changes. Signal contains a
        reference to the :class:`~pyqtgraph.ViewBox`, and tuple of (xmin, xmax).
    sigXRangeChanged : Signal
        Signal is emitted when the range on the x-axis changes. Signal contains a
        reference to the :class:`~pyqtgraph.ViewBox`, and tuple of (ymin, ymax).
    sigRangeChanged : Signal
        Signal is emitted when the range on either x or y-axis changes. Signal contains
        a reference to the :class:`~pyqtgraph.ViewBox`, a list of lists of the form,
        ``[[x_min, x_max], [y_min, y_max]]`` and a list of two booleans, indicating if
        the range on which axis has changed of the form 
        ``[x_range_changed, y_range_changed]``.
    """
    
    sigRangeChanged = QtCore.Signal(object, object)    ## Emitted when the ViewBox range has changed
    sigYRangeChanged = QtCore.Signal(object, object)   ## Emitted when the ViewBox Y range has changed
    sigXRangeChanged = QtCore.Signal(object, object)   ## Emitted when the ViewBox X range has changed
    sigRangeChangedManually = QtCore.Signal(object)    ## Emitted when the ViewBox range is changed via user interaction
        
    lastFileDir = None
    
    def __init__(
        self,
        parent : QtCore.QObject | None = None,
        name: str | None = None,
        labels: dict[str, str] | None = None,
        title: str | None = None,
        viewBox: ViewBox | None = None,
        axisItems: dict[str, AxisItem] | None = None,
        enableMenu: bool = True,
        **kwargs
    ):  
        super().__init__(parent)

        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Expanding
        )

        ## Set up control buttons
        self.autoBtn = ButtonItem(icons.getGraphPixmap('auto'), 14, self)
        self.autoBtn.mode = 'auto'
        self.autoBtn.clicked.connect(self.autoBtnClicked)
        self.buttonsHidden = False  # has the user has requested buttons to be hidden?
        self.mouseHovering = False

        self.layout = QtWidgets.QGraphicsGridLayout()
        self.layout.setContentsMargins(1,1,1,1)
        self.setLayout(self.layout)
        self.layout.setHorizontalSpacing(0)
        self.layout.setVerticalSpacing(0)

        if viewBox is None:
            viewBox = ViewBox(parent=self, enableMenu=enableMenu)
        self.vb = viewBox
        self.vb.sigStateChanged.connect(self.viewStateChanged)

        # Enable or disable plotItem menu
        self.setMenuEnabled(enableMenu, None)

        if name is not None:
            self.vb.register(name)
        self.vb.sigRangeChanged.connect(self.sigRangeChanged)
        self.vb.sigXRangeChanged.connect(self.sigXRangeChanged)
        self.vb.sigYRangeChanged.connect(self.sigYRangeChanged)
        self.vb.sigRangeChangedManually.connect(self.sigRangeChangedManually)

        self.layout.addItem(self.vb, 2, 1)
        self.alpha = 1.0
        self.autoAlpha = True
        self.spectrumMode = False

        self.legend = None

        # Initialize axis items
        self.axes = {}
        self.setAxisItems(axisItems)

        self.titleLabel = LabelItem('', size='11pt', parent=self)
        self.layout.addItem(self.titleLabel, 0, 1)
        self.setTitle(None)  ## hide

        for i in range(4):
            self.layout.setRowPreferredHeight(i, 0)
            self.layout.setRowMinimumHeight(i, 0)
            self.layout.setRowSpacing(i, 0)
            self.layout.setRowStretchFactor(i, 1)

        for i in range(3):
            self.layout.setColumnPreferredWidth(i, 0)
            self.layout.setColumnMinimumWidth(i, 0)
            self.layout.setColumnSpacing(i, 0)
            self.layout.setColumnStretchFactor(i, 1)
        self.layout.setRowStretchFactor(2, 100)
        self.layout.setColumnStretchFactor(1, 100)


        self.items = []
        self.curves = []
        self.itemMeta = weakref.WeakKeyDictionary()
        self.dataItems = []
        self.paramList = {}
        self.avgCurves = {}
        # Change these properties to adjust the appearance of the averaged curve:
        self.avgPen = fn.mkPen((0, 200, 0))
        self.avgShadowPen = fn.mkPen((0, 0, 0), width=4)

        ### Set up context menu

        w = QtWidgets.QWidget()
        self.ctrl = c = ui_template.Ui_Form()
        c.setupUi(w)

        menuItems = [
            (translate("PlotItem", 'Transforms'), c.transformGroup),
            (translate("PlotItem", 'Downsample'), c.decimateGroup),
            (translate("PlotItem", 'Average'), c.averageGroup),
            (translate("PlotItem", 'Alpha'), c.alphaGroup),
            (translate("PlotItem", 'Grid'), c.gridGroup),
            (translate("PlotItem", 'Points'), c.pointsGroup),
        ]

        self.ctrlMenu = QtWidgets.QMenu(translate("PlotItem", 'Plot Options'))

        for name, grp in menuItems:
            sm = self.ctrlMenu.addMenu(name)
            act = QtWidgets.QWidgetAction(self)
            act.setDefaultWidget(grp)
            sm.addAction(act)

        self.stateGroup = WidgetGroup()
        for name, w in menuItems:
            self.stateGroup.autoAdd(w)

        self.fileDialog = None

        c.alphaGroup.toggled.connect(self.updateAlpha)
        c.alphaSlider.valueChanged.connect(self.updateAlpha)
        c.autoAlphaCheck.toggled.connect(self.updateAlpha)

        c.xGridCheck.toggled.connect(self.updateGrid)
        c.yGridCheck.toggled.connect(self.updateGrid)
        c.gridAlphaSlider.valueChanged.connect(self.updateGrid)

        c.fftCheck.toggled.connect(self.updateSpectrumMode)
        c.subtractMeanCheck.toggled.connect(self.updateSubtractMeanMode)
        c.logXCheck.toggled.connect(self.updateLogMode)
        c.logYCheck.toggled.connect(self.updateLogMode)
        c.derivativeCheck.toggled.connect(self.updateDerivativeMode)
        c.phasemapCheck.toggled.connect(self.updatePhasemapMode)

        c.downsampleSpin.valueChanged.connect(self.updateDownsampling)
        c.downsampleCheck.toggled.connect(self.updateDownsampling)
        c.autoDownsampleCheck.toggled.connect(self.updateDownsampling)
        c.subsampleRadio.toggled.connect(self.updateDownsampling)
        c.meanRadio.toggled.connect(self.updateDownsampling)
        c.clipToViewCheck.toggled.connect(self.updateDownsampling)

        self.ctrl.avgParamList.itemClicked.connect(self.avgParamListClicked)
        self.ctrl.averageGroup.toggled.connect(self.avgToggled)

        self.ctrl.maxTracesCheck.toggled.connect(self._handle_max_traces_toggle)
        self.ctrl.forgetTracesCheck.toggled.connect(self.updateDecimation)
        self.ctrl.maxTracesSpin.valueChanged.connect(self.updateDecimation)

        if labels is None:
            labels = {}
        for label in list(self.axes.keys()):
            if label in kwargs:
                labels[label] = kwargs[label]
                del kwargs[label]
        for k in labels:
            if isinstance(labels[k], str):
                labels[k] = (labels[k],)
            self.setLabel(k, *labels[k])

        if title is not None:
            self.setTitle(title)

        if kwargs:
            self.plot(**kwargs)        
        
    def implements(self, interface=None):
        return interface in ['ViewBoxWrapper']

    def getViewBox(self) -> ViewBox:
        """
        Return the embedded :class:`~pyqtgraph.ViewBox`.

        Returns
        -------
        ViewBox 
            The :class:`ViewBox <pyqtgraph.ViewBox>` that this PlotItem uses.
        """
        return self.vb
    
    # Wrap a few methods from viewBox. 
    # Important: don't use settattr(m, getattr(self.vb, m)) as we'd be leaving the
    # viewbox alive because we had a reference to an instance method (creating wrapper
    # methods at runtime instead).

    # Note: If you update this list, please update the class docstring as well
    for m in ['setXRange', 'setYRange', 'setXLink', 'setYLink', 'setAutoPan',
              'setAutoVisible', 'setDefaultPadding', 'setRange', 'autoRange',
              'viewRect', 'viewRange', 'setMouseEnabled', 'setLimits',
              'enableAutoRange', 'disableAutoRange', 'setAspectLocked', 'invertY',
              'invertX', 'register', 'unregister']:
                
        def _create_method(name):
            def method(self, *args, **kwargs):
                return getattr(self.vb, name)(*args, **kwargs)
            method.__name__ = name
            return method
        
        locals()[m] = _create_method(m)
        
    del _create_method
    
    def setAxisItems(self, axisItems: 'dict[str, AxisItem] | None' = None) -> None:
        """
        Place axis items and initializes non-existing axis items.

        Parameters
        ----------
        axisItems : dict of str and AxisItem or None
            Optional dictionary instructing the PlotItem to use pre-constructed items
            for its axes. The dict keys must be axis names
            {'left', 'bottom', 'right', 'top'} and the values must be instances of
            :class:`~pyqtgraph.AxisItem`.
        """
        
        if axisItems is None:
            axisItems = {}
        
        # Array containing visible axis items
        # Also containing potentially hidden axes, but they are not touched so it does
        # not matter
        visibleAxes = ['left', 'bottom']
        # Note that it does not matter that this adds
        # some values to visibleAxes a second time
        visibleAxes.extend(axisItems.keys()) 
        
        for k, pos in (('top', (1,1)), ('bottom', (3,1)), ('left', (2,0)), ('right', (2,2))):
            if k in self.axes:
                if k not in axisItems:
                    continue # Nothing to do here
                
                # Remove old axis
                oldAxis = self.axes[k]['item']
                self.layout.removeItem(oldAxis)
                if oldAxis.scene() is not None:
                    oldAxis.scene().removeItem(oldAxis)
                oldAxis.unlinkFromView()
            
            # Create new axis
            if k in axisItems:
                axis = axisItems[k]
                if (
                    axis.scene() is not None 
                    and (k not in self.axes or axis != self.axes[k]["item"])
                ):
                    raise RuntimeError(
                        "Can't add an axis to multiple plots. Shared axes "
                        "can be achieved with multiple AxisItem instances "
                        "and set[X/Y]Link."
                    )
            else:
                axis = AxisItem(orientation=k, parent=self)
            
            # Set up new axis
            axis.linkToView(self.vb)
            self.axes[k] = {'item': axis, 'pos': pos}
            self.layout.addItem(axis, *pos)
            # place axis above images at z=0, items that want to draw over the axes 
            # should be placed at z>=1
            axis.setZValue(0.5) 
            axis.setFlag(
                QtWidgets.QGraphicsItem.GraphicsItemFlag.ItemNegativeZStacksBehindParent
            )           
            axisVisible = k in visibleAxes
            self.showAxis(k, axisVisible)
        
    def setLogMode(self, x=None, y=None):
        """
        Set log scaling for `x` and/or `y` axes.

        This informs PlotDataItems to transform logarithmically and switches the axes to
        use log ticking.

        Parameters
        ----------
        x : bool or None
            If ``True``, set the x-axis to log mode. A value of ``None`` is ignored.
        y : bool or None
            If ``True``, set the y-axis to log mode. A value of ``None`` is ignored.
        
        Notes
        -----
        No other items, in the scene will be affected by this; there is (currently) 
        no generic way to redisplay a :class:`~pyqtgraph.GraphicsItem` with log
        coordinates.
        """
        if x is not None:
            self.ctrl.logXCheck.setChecked(x)
        if y is not None:
            self.ctrl.logYCheck.setChecked(y)
        
    def showGrid(self, x=None, y=None, alpha=None):
        """
        Show or hide the grid for either axis.

        Parameters
        ----------
        x : bool or None
            Show the X grid, a value of ``None`` is ignored.
        y : bool or None
            Show the Y grid, a value of ``None`` is ignored.
        alpha : float or None
            Opacity of the grid, float values need to be between [0.0, 1.0].
        """
        if x is None and y is None and alpha is None:
            # prevent people getting confused if they just call showGrid()
            raise ValueError("Must specify at least one of x, y, or alpha.")
        
        if x is not None:
            self.ctrl.xGridCheck.setChecked(x)
        if y is not None:
            self.ctrl.yGridCheck.setChecked(y)
        if alpha is not None:
            v = fn.clip_scalar(alpha, 0, 1) * self.ctrl.gridAlphaSlider.maximum() # slider range 0 to 255
            self.ctrl.gridAlphaSlider.setValue( int(v) )
        
    def close(self):
        ## Most of this crap is needed to avoid PySide trouble. 
        ## The problem seems to be whenever scene.clear() leads to deletion of widgets (either through proxies or qgraphicswidgets)
        ## the solution is to manually remove all widgets before scene.clear() is called
        if self.ctrlMenu is None: ## already shut down
            return
        self.ctrlMenu.setParent(None)
        self.ctrlMenu = None
        
        self.autoBtn.setParent(None)
        self.autoBtn = None
        
        for k in self.axes:
            i = self.axes[k]['item']
            i.close()
            
        self.axes = None
        self.scene().removeItem(self.vb)
        self.vb = None
        
    def registerPlot(self, name):   ## for backward compatibility
        self.vb.register(name)
        
    @QtCore.Slot(bool)
    @QtCore.Slot(int)
    def updateGrid(self, *args):
        alpha = self.ctrl.gridAlphaSlider.value()
        x = alpha if self.ctrl.xGridCheck.isChecked() else False
        y = alpha if self.ctrl.yGridCheck.isChecked() else False
        self.getAxis('top').setGrid(x)
        self.getAxis('bottom').setGrid(x)
        self.getAxis('left').setGrid(y)
        self.getAxis('right').setGrid(y)

    def viewGeometry(self) -> QtCore.QRectF:
        """
        Return the screen geometry of the viewbox.

        Returns
        -------
        QRectF
            The QRectF instance that contains the view area.
        """
        v = self.scene().views()[0]
        b = self.vb.mapRectToScene(self.vb.boundingRect())
        wr = v.mapFromScene(b).boundingRect()
        pos = v.mapToGlobal(v.pos())
        wr.adjust(pos.x(), pos.y(), pos.x(), pos.y())
        return wr

    @QtCore.Slot(bool)
    def avgToggled(self, b: bool):
        if b:
            self.recomputeAverages()
        for k in self.avgCurves:
            self.avgCurves[k][1].setVisible(b)
        
    @QtCore.Slot(QtWidgets.QListWidgetItem)
    def avgParamListClicked(self, item):
        name = str(item.text())
        self.paramList[name] = (item.checkState() == QtCore.Qt.CheckState.Checked)
        self.recomputeAverages()
        
    def recomputeAverages(self):
        if not self.ctrl.averageGroup.isChecked():
            return
        for k in self.avgCurves:
            self.removeItem(self.avgCurves[k][1])
        self.avgCurves = {}
        for c in self.curves:
            self.addAvgCurve(c)
        self.replot()
        
    def addAvgCurve(self, curve):
        ## Add a single curve into the pool of curves averaged together

        ## If there are plot parameters, then we need to determine which to average together.
        remKeys = []
        addKeys = []
        if self.ctrl.avgParamList.count() > 0:

            # First determine the key of the curve to which this new data should be averaged
            for i in range(self.ctrl.avgParamList.count()):
                item = self.ctrl.avgParamList.item(i)
                if item.checkState() == QtCore.Qt.CheckState.Checked:
                    remKeys.append(str(item.text()))
                else:
                    addKeys.append(str(item.text()))

            if not remKeys:
                # In this case,there would be 1 avg plot for each data plot; not useful
                return

        p = self.itemMeta.get(curve,{}).copy()
        for k in p:
            if type(k) is tuple:
                p['.'.join(k)] = p[k]
                del p[k]
        for rk in remKeys:
            if rk in p:
                del p[rk]
        for ak in addKeys:
            if ak not in p:
                p[ak] = None
        key = tuple(p.items())

        ### Create a new curve if needed
        if key not in self.avgCurves:
            plot = PlotDataItem()
            plot.setPen( self.avgPen )
            plot.setShadowPen(  self.avgShadowPen )
            plot.setAlpha(1.0, False)
            plot.setZValue(100)
            self.addItem(plot, skipAverage=True)
            self.avgCurves[key] = [0, plot]
        self.avgCurves[key][0] += 1
        (n, plot) = self.avgCurves[key]

        ### Average data together
        (x, y) = curve.getData()
        stepMode = curve.opts['stepMode']
        if plot.yData is not None and y.shape == plot.yData.shape:
            # note that if shapes do not match, then the average resets.
            newData = plot.yData * (n-1) / float(n) + y * 1.0 / float(n)
            plot.setData(plot.xData, newData, stepMode=stepMode)
        else:
            plot.setData(x, y, stepMode=stepMode)
        
    @QtCore.Slot()
    def autoBtnClicked(self):
        if self.autoBtn.mode == 'auto':
            self.enableAutoRange()
            self.autoBtn.hide()
            self.sigRangeChangedManually.emit(self.vb.mouseEnabled()[:])
        else:
            self.disableAutoRange()
            
    @QtCore.Slot()
    def viewStateChanged(self):
        self.updateButtons()

    def addItem(self, item, *args, **kwargs):
        """
        Add a :class:`~pyqtgraph.GraphicsItem` to the :class:`~pyqtgraph.ViewBox`.
        
        If the item has plot data (:class:`PlotDataItem <pyqtgraph.PlotDataItem>` , 
        :class:`~pyqtgraph.PlotCurveItem`, :class:`~pyqtgraph.ScatterPlotItem` ), it may
        be included in analysis performed by the PlotItem.

        Parameters
        ----------
        item : GraphicsItem
            Item to add to the ViewBox.
        *args : tuple
            Arguments relayed to :meth:`~pyqtgraph.ViewBox.addItem`.
        **kwargs : dict
            Keyword arguments for adding an item.  Supported arguments include.

            ============ ===============================================================
            Property     Description
            ============ ===============================================================
            ignoreBounds ``bool`` - Keyword argument relayed to 
                         :meth:`~pyqtgraph.ViewBox.addItem`.
            
            skipAverage  ``bool`` - If ``True``, do not use curve to compute average
                         curves.
            ============ ===============================================================

        See Also
        --------
        :meth:`~pyqtgraph.ViewBox.addItem`
            See method for supported arguments to be passed.
        """
        if item in self.items:
            warnings.warn('Item already added to PlotItem, ignoring.')
            return
        self.items.append(item)
        vbargs = {}
        if 'ignoreBounds' in kwargs:
            vbargs['ignoreBounds'] = kwargs['ignoreBounds']
        self.vb.addItem(item, *args, **vbargs)
        name = None
        if hasattr(item, 'implements') and item.implements('plotData'):
            name = item.name()
            self.dataItems.append(item)            
            params = kwargs.get('params', {})
            self.itemMeta[item] = params
            self.curves.append(item)
        
        # Toggle log mode if item implements setLogMode and selected in the context menu
        if hasattr(item, 'setLogMode'):
            item.setLogMode(
                self.ctrl.logXCheck.isChecked(),
                self.ctrl.logYCheck.isChecked()
            )
            
        if isinstance(item, PlotDataItem):
            ## configure curve for this plot
            (alpha, auto) = self.alphaState()
            item.setAlpha(alpha, auto)
            item.setSubtractMeanMode(self.ctrl.subtractMeanCheck.isChecked())
            item.setFftMode(self.ctrl.fftCheck.isChecked())
            item.setDownsampling(*self.downsampleMode())
            item.setClipToView(self.clipToViewMode())
            
            ## Hide older plots if needed
            self.updateDecimation()
            
            ## Add to average if needed
            self.updateParamList()
            if self.ctrl.averageGroup.isChecked() and 'skipAverage' not in kwargs:
                self.addAvgCurve(item)

        # conditionally add item to the legend
        if name is not None and hasattr(self, 'legend') and self.legend is not None:
            self.legend.addItem(item, name=name)            

    def listDataItems(self):
        """
        Return a list of all data items current plotted.

        Returns
        -------
        list of PlotCurveItem, ScatterPlotItem or PlotDataItem
            A copy of a list of the data items.
        """
        return self.dataItems[:]

    def addLine(self, x=None, y=None, z=None, **kwargs):
        """
        Create a new :class:`~pyqtgraph.InfiniteLine` and add it to the plot.

        Parameters
        ----------
        x : float or None
            Position in the x-axis to draw the line. If specified, the line will be 
            vertical. Default ``None``.
        y : float or None
            Position in the y-axis to draw the line. If specified, the line will be
            horizontal. Default ``None``.
        z : int or None
            Z value to set the line to. This is used to determine which items are on top
            of the other.  See :meth:`~QtWidgets.QGraphicsItem.setZValue`.
        **kwargs : dict
            Keyword arguments to pass to the :class:`~pyqtgraph.InfiniteLine` instance.
        
        Returns
        -------
        InfiniteLine
            The new :class:`~pyqtgraph.InfiniteLine` added.
        """
        kwargs['pos'] = kwargs.get('pos', x if x is not None else y)
        kwargs['angle'] = kwargs.get('angle', 0 if x is None else 90)
        line = InfiniteLine(**kwargs)
        self.addItem(line)
        if z is not None:
            line.setZValue(z)
        return line

    def removeItem(self, item):
        """
        Remove an item from the plot.

        Parameters
        ----------
        item : GraphicsItem
            The item to remove.
        """
        if item not in self.items:
            return
        self.items.remove(item)
        if item in self.dataItems:
            self.dataItems.remove(item)

        self.vb.removeItem(item)

        if item in self.curves:
            self.curves.remove(item)
            self.updateDecimation()
            self.updateParamList()

        if self.legend is not None:
            self.legend.removeItem(item)

    def clear(self):
        """
        Remove all items from the PlotItem's :class:`~pyqtgraph.ViewBox`.
        """
        for i in self.items[:]:
            self.removeItem(i)
        self.avgCurves = {}
    
    def clearPlots(self):
        """
        Remove all curves from the PlotItem's :class:`~pyqtgraph.ViewBox`.
        """
        for i in self.curves[:]:
            self.removeItem(i)
        self.avgCurves = {}        
    
    def plot(self, *args, **kwargs):
        """
        Add and return a new :class:`~pyqtgraph.PlotDataItem`.

        Parameters
        ----------
        *args : tuple, optional
            Arguments that are passed to the :class:`~pyqtgraph.PlotDataItem`
            constructor.
        **kwargs : dict, optional
            Keyword arguments that are passed to the :class:`~pyqtgraph.PlotDataItem`
            constructor.  In addition, the following keyword arguments are accepted.

            =========== ================================================================
            Property    Description
            =========== ================================================================
            clear       ``bool`` - Call :meth:`~pyqtgraph.PlotItem.clear` prior to
                        creating the new plot instance.
            
            params      ``dict`` or ``None`` - Arguments to passed as the `params`
                        argument to :meth:`~pyqtgraph.PlotItem.addItem`.
            =========== ================================================================

        Returns
        -------
        PlotDataItem
            The newly created :class:`~pyqtgraph.PlotDataItem`.
        """
        clear = kwargs.get('clear', False)
        params = kwargs.get('params')
          
        if clear:
            self.clear()
            
        item = PlotDataItem(*args, **kwargs)
            
        if params is None:
            params = {}
        self.addItem(item, params=params)
        
        return item

    def addLegend(self, offset=(30, 30), **kwargs):
        """
        Add a new :class:`~pyqtgraph.LegendItem`.
         
        After the LegendItem is created, it is anchored in the internal 
        :class:`~pyqtgraph.ViewBox`. Plots added after this will be automatically 
        displayed in the legend if they are created with a 'name' argument.

        If a :class:`~pyqtgraph.LegendItem` has already been created using this method, 
        that item will be returned rather than creating a new one.

        Parameters
        ----------
        offset : tuple of int, int
            The distance to offset the LegendItem, defaults to ``(30, 30)``.
        **kwargs : dict, optional
            Keyword argument passed to the :class:`~pyqtgraph.LegendItem` constructor.
        
        Returns
        -------
        LegendItem
            The instance of :class:`~pyqtgraph.LegendItem` that was constructed.
        """
        if self.legend is None:
            self.legend = LegendItem(offset=offset, **kwargs)
            self.legend.setParentItem(self.vb)
        return self.legend
        
    def addColorBar(self, image, **kwargs):
        """
        Add a ColorBarItem and set to the provided image.
        
        A call like ``plot.addColorBar(img, colorMap='viridis')`` is a convenient
        method to assign and show a color map.

        Parameters
        ----------
        image : ImageItem or list of ImageItem
            See :meth:`~pyqtgraph.ColorBarItem.setImageItem` for details.
        **kwargs : dict, optional
            Keyword arguments passed to the :class:`~pyqtgraph.ColorBarItem`
            constructor.
        
        Returns
        -------
        ColorBarItem
            The newly created :class:`~pyqtgraph.ColorBarItem` instance.
        """
        from ..ColorBarItem import ColorBarItem  # avoid circular import
        bar = ColorBarItem(**kwargs)
        bar.setImageItem( image, insert_in=self )
        return bar

    def multiDataPlot(self, *, x=None, y=None, constKwargs=None, **kwargs):
        """
        Allow plotting multiple curves on the same plot in one call.

        Parameters
        ----------
        x, y : array_like
            Can be in the following formats:
              - {x or y} = [n1, n2, n3, ...]: The named argument iterates through
                ``n`` curves, while the unspecified argument is range(len(n)) for
                each curve.
              - x, [y1, y2, y3, ...]
              - [x1, x2, x3, ...], [y1, y2, y3, ...]
              - [x1, x2, x3, ...], y

            where ``x_n`` and ``y_n`` are ``ndarray`` data for each curve. Since
            ``x`` and ``y`` values are matched using ``zip``, unequal lengths mean
            the longer array will be truncated. Note that 2D matrices for either x
            or y are considered lists of curve data.
        constKwargs : dict, optional
            A dict of {str: value} passed to each curve during ``plot()``.
        **kwargs : dict, optional
            A dict of {str: iterable} where the str is the name of a kwarg and the
            iterable is a list of values, one for each plotted curve.
        
        Returns
        -------
        list of PlotDataItem
            Returns a list of the newly constructed :class:`~pyqtgraph.PlotDataItem`
            instances representing the new curves.
        """
        if (x is not None and not len(x)) or (y is not None and not len(y)):
            # Nothing to plot -- either x or y array will bail out early from
            # zip() below.
            return []
        def scalarOrNone(val):
            return val is None or (len(val) and np.isscalar(val[0]))

        if scalarOrNone(x) and scalarOrNone(y):
            raise ValueError(
                "If both `x` and `y` represent single curves, use `plot` instead "
                "of `multiPlot`."
            )
        curves = []
        constKwargs = constKwargs or {}
        xy: dict[str, list | None] = dict(x=x, y=y)
        for key, oppositeVal in zip(('x', 'y'), [y, x]):
            oppositeVal: Iterable | None
            val = xy[key]
            if val is None:
                # Other curve has all data, make range that supports longest chain
                val = range(max(len(curveN) for curveN in oppositeVal))
            if np.isscalar(val[0]):
                # x, [y1, y2, y3, ...] or [x1, x2, x3, ...], y
                # Repeat the single curve to match length of opposite list
                val = [val] * len(oppositeVal)
            xy[key] = val
        for ii, (xi, yi) in enumerate(zip(xy['x'], xy['y'])):
            for kk in kwargs:
                if len(kwargs[kk]) <= ii:
                    raise ValueError(
                        f"Not enough values for kwarg `{kk}`. "
                        f"Expected {ii + 1:d} (number of curves to plot), got"
                        f" {len(kwargs[kk]):d}"
                    )
            kwargs_i = {kk: kwargs[kk][ii] for kk in kwargs}
            constKwargs.update(kwargs_i)
            curves.append(self.plot(xi, yi, **constKwargs))
        return curves

    def scatterPlot(self, *args, **kwargs):
        """
        Create a :class:`~pyqtgraph.PlotDataItem` and add it to the plot.

        The :class:`~pyqtgraph.PlotDataItem` instance will render the underlying data
        in a scatter plot form.

        Parameters
        ----------
        *args : tuple, optional
            Arguments that are passed to the :class:`~pyqtgraph.PlotDataItem`
            constructor.
        **kwargs : dict, optional
            Keyword arguments that are passed to the :class:`~pyqtgraph.PlotDataItem`
            constructor.  In addition, the following keyword arguments are accepted.

            =========== ================================================================
            Property    Description
            =========== ================================================================
            pen         ``QPen`` - Sets the pen used to draw lines or symbols for the
                        plot.  Equivalent to ``symbolPen``.

            brush       ``QBrush`` - Sets the brush used to fill symbols for the plot.
                        Equivalent to ``symbolBrush``.

            size        ``float`` - Sets the symbol size for the plot. Equivalent to
                        ``symbolSize``.
            =========== ================================================================

        Returns
        -------
        PlotDataItem
            The newly created :class:`~pyqtgraph.PlotDataItem`.
        """
        if 'pen' in kwargs:
            kwargs['symbolPen'] = kwargs['pen']
        kwargs['pen'] = None
            
        if 'brush' in kwargs:
            kwargs['symbolBrush'] = kwargs['brush']
            del kwargs['brush']
            
        if 'size' in kwargs:
            kwargs['symbolSize'] = kwargs['size']
            del kwargs['size']

        return self.plot(*args, **kwargs)
                
    def replot(self):
        self.update()

    def updateParamList(self):
        self.ctrl.avgParamList.clear()
        ## Check to see that each parameter for each curve is present in the list
        for c in self.curves:
            for p in list(self.itemMeta.get(c, {}).keys()):
                if type(p) is tuple:
                    p = '.'.join(p)

                if matches := self.ctrl.avgParamList.findItems(
                    p, QtCore.Qt.MatchFlag.MatchExactly
                ):
                    i = matches[0]
                else:
                    i = QtWidgets.QListWidgetItem(p)
                    if p in self.paramList and self.paramList[p] is True:
                        i.setCheckState(QtCore.Qt.CheckState.Checked)
                    else:
                        i.setCheckState(QtCore.Qt.CheckState.Unchecked)
                    self.ctrl.avgParamList.addItem(i)
                self.paramList[p] = (i.checkState() == QtCore.Qt.CheckState.Checked)

    def writeSvg(self, fileName=None):
        """
        Write the plot content to an SVG file.

        Parameters
        ----------
        fileName : str, optional
            The name of the file to write to. If not specified, a file dialog will be
            opened.
        """
        if fileName is None:
            self._chooseFilenameDialog(handler=self.writeSvg)
            return

        fileName = str(fileName)
        PlotItem.lastFileDir = os.path.dirname(fileName)
        
        from ...exporters import SVGExporter
        ex = SVGExporter(self)
        ex.export(fileName)
        
    def writeImage(self, fileName=None):
        """
        Write the plot content to an image file.

        Parameters
        ----------
        fileName : str, optional
            The name of the file to write to. If not specified, a file dialog will be
            opened.
        """
        if fileName is None:
            self._chooseFilenameDialog(handler=self.writeImage)
            return

        from ...exporters import ImageExporter
        ex = ImageExporter(self)
        ex.export(fileName)
        
    def writeCsv(self, fileName=None):
        """
        Write the plot data to a CSV file.

        Parameters
        ----------
        fileName : str, optional
            The name of the file to write to. If not specified, a file dialog will be
            opened.
        """
        if fileName is None:
            self._chooseFilenameDialog(handler=self.writeCsv)
            return

        fileName = str(fileName)
        PlotItem.lastFileDir = os.path.dirname(fileName)
        
        data = [c.getData() for c in self.curves]
        with open(fileName, 'w') as fd:
            i = 0
            while True:
                done = True
                for d in data:
                    if i < len(d[0]):
                        fd.write('%g,%g,' % (d[0][i], d[1][i]))
                        done = False
                    else:
                        fd.write(' , ,')
                fd.write('\n')
                if done:
                    break
                i += 1

    def saveState(self):
        state = self.stateGroup.state()
        state['paramList'] = self.paramList.copy()
        state['view'] = self.vb.getState()
        return state
        
    def restoreState(self, state):
        if 'paramList' in state:
            self.paramList = state['paramList'].copy()
            
        self.stateGroup.setState(state)
        self.updateSpectrumMode()
        self.updateSubtractMeanMode()
        self.updateDownsampling()
        self.updateAlpha()
        self.updateDecimation()
        
        if 'powerSpectrumGroup' in state:
            state['fftCheck'] = state['powerSpectrumGroup']
        if 'gridGroup' in state:
            state['xGridCheck'] = state['gridGroup']
            state['yGridCheck'] = state['gridGroup']
            
        self.stateGroup.setState(state)
        self.updateParamList()
        
        if 'view' not in state:
            r = [[float(state['xMinText']), float(state['xMaxText'])], [float(state['yMinText']), float(state['yMaxText'])]]
            state['view'] = {
                'autoRange': [state['xAutoRadio'], state['yAutoRadio']],
                'linkedViews': [state['xLinkCombo'], state['yLinkCombo']],
                'targetRange': r,
                'viewRange': r,
            }
        self.vb.setState(state['view'])

    def widgetGroupInterface(self):
        return (None, PlotItem.saveState, PlotItem.restoreState)
      
    @QtCore.Slot(bool)
    def updateSpectrumMode(self, b=None):
        if b is None:
            b = self.ctrl.fftCheck.isChecked()
        for c in self.curves:
            c.setFftMode(b)
        self.enableAutoRange()
        self.recomputeAverages()

    @QtCore.Slot()
    def updateSubtractMeanMode(self):
        d = self.ctrl.subtractMeanCheck.isChecked()
        for i in self.items:
            if hasattr(i, 'setSubtractMeanMode'):
                i.setSubtractMeanMode(d)
        self.enableAutoRange()
        self.recomputeAverages()

    @QtCore.Slot()
    def updateLogMode(self):
        x = self.ctrl.logXCheck.isChecked()
        y = self.ctrl.logYCheck.isChecked()
        for i in self.items:
            if hasattr(i, 'setLogMode'):
                i.setLogMode(x,y)
        self.getAxis('bottom').setLogMode(x, y)
        self.getAxis('top').setLogMode(x, y)
        self.getAxis('left').setLogMode(x, y)
        self.getAxis('right').setLogMode(x, y)
        self.enableAutoRange()
        self.recomputeAverages()
    
    @QtCore.Slot()
    def updateDerivativeMode(self):
        d = self.ctrl.derivativeCheck.isChecked()
        for i in self.items:
            if hasattr(i, 'setDerivativeMode'):
                i.setDerivativeMode(d)
        self.enableAutoRange()
        self.recomputeAverages()

    @QtCore.Slot()
    def updatePhasemapMode(self):
        d = self.ctrl.phasemapCheck.isChecked()
        for i in self.items:
            if hasattr(i, 'setPhasemapMode'):
                i.setPhasemapMode(d)
        self.enableAutoRange()
        self.recomputeAverages()
        
        
    def setDownsampling(self, ds=None, auto=None, mode=None):
        """
        Set the downsampling mode for the PlotItem.

        Downsampling reduces the number of samples drawn to improve performance.
        The downsampling mode can be set to 'subsample', 'mean', 'peak', or None.
        If downsampling is enabled, the view will display downsampled data when zoomed out,
        but will display original data at high zoom levels.

        Parameters
        ----------
        ds : int or bool or None, optional
            The downsampling factor. If ``None``, the downsampling factor is not
            changed. If ``True``, the downsampling factor is set to the value in the
            downsampling spin box. If ``False``, downsampling is disabled. If an
            integer, the downsampling factor is set to this value. Default is ``None``.
        auto : bool or None, optional
            If ``True``, automatic downsampling is enabled. If ``False``, automatic
            downsampling is disabled. If ``None``, the automatic downsampling setting
            is not changed. Default is ``None``.
        mode : {'subsample', 'mean', 'peak'} or None, optional
            The downsampling mode. If ``None``, the downsampling mode is not changed.
            Default is ``None``.
        
        Raises
        ------
        ValueError
            Raised if the specified downsample mode is not recognized.
        """

        if ds is not None:
            if ds is False:
                self.ctrl.downsampleCheck.setChecked(False)
            elif ds is True:
                self.ctrl.downsampleCheck.setChecked(True)
            else:
                self.ctrl.downsampleCheck.setChecked(True)
                self.ctrl.downsampleSpin.setValue(ds)
                
        if auto is not None:
            if auto and ds is not False:
                self.ctrl.downsampleCheck.setChecked(True)
            self.ctrl.autoDownsampleCheck.setChecked(auto)
            
        if mode is not None:
            if mode == 'subsample':
                self.ctrl.subsampleRadio.setChecked(True)
            elif mode == 'mean':
                self.ctrl.meanRadio.setChecked(True)
            elif mode == 'peak':
                self.ctrl.peakRadio.setChecked(True)
            else:
                raise ValueError(
                    "mode argument must be 'subsample', 'mean', or 'peak'."
                )
            
    @QtCore.Slot()
    def updateDownsampling(self):
        ds, auto, method = self.downsampleMode()
        clip = self.ctrl.clipToViewCheck.isChecked()
        for c in self.curves:
            c.setDownsampling(ds, auto, method)
            c.setClipToView(clip)
        self.recomputeAverages()
        
    def downsampleMode(self) -> tuple[int, bool, str]:
        if self.ctrl.downsampleCheck.isChecked():
            ds = self.ctrl.downsampleSpin.value()
        else:
            ds = 1
            
        auto = (
            self.ctrl.downsampleCheck.isChecked() and 
            self.ctrl.autoDownsampleCheck.isChecked()
        )
            
        if self.ctrl.subsampleRadio.isChecked():
            method = 'subsample' 
        elif self.ctrl.meanRadio.isChecked():
            method = 'mean'
        elif self.ctrl.peakRadio.isChecked():
            method = 'peak'
        else:
            raise ValueError(
                "One of the method radios must be selected for: 'subsample', 'mean', "
                "'peak'."
            )
        return ds, auto, method
        
    def setClipToView(self, clip):
        """
        Set the default clip-to-view mode for new curves.

        Parameters
        ----------
        clip : bool
            If ``True``, new curves will be clipped to the view box.
        """
        self.ctrl.clipToViewCheck.setChecked(clip)
        
    def clipToViewMode(self) -> bool:
        """
        Return whether clip-to-view mode is enabled.

        Returns
        -------
        bool
            ``True`` if clip-to-view mode is enabled, ``False`` otherwise.
        """
        
        return self.ctrl.clipToViewCheck.isChecked()
    
    @QtCore.Slot(bool)
    def _handle_max_traces_toggle(self, check_state):
        if check_state:
            self.updateDecimation()
        else:
            for curve in self.curves:
                curve.show()
    
    @QtCore.Slot()
    def updateDecimation(self):
        """
        Update the number of visible curves.

        Reduce or increase number of visible curves according to value set by the 
        `Max Traces` spinner, if `Max Traces` is checked in the context menu. Destroy
        curves that are not visible if `forget traces` is checked. In most cases, this
        function is called automatically when the `Max Traces` GUI elements are
        triggered. It is also called when the state of :class:`~pyqtgraph.PlotItem` is
        updated, its state is restored, or new items added added/removed.
        
        This can cause an unexpected or conflicting state of curve visibility
        (or destruction) if curve visibilities are controlled externally. In the case of
        external control it is advised to disable the `Max Traces` checkbox (or context
        menu) to prevent unexpected curve state changes.

        :meta private:
        """
        if not self.ctrl.maxTracesCheck.isChecked():
            return
        else:
            numCurves = self.ctrl.maxTracesSpin.value()

        if self.ctrl.forgetTracesCheck.isChecked():
            for curve in self.curves[:-numCurves]:
                curve.clear()
                self.removeItem(curve)

        for i, curve in enumerate(reversed(self.curves)):
            if i < numCurves:
                curve.show()
            else:
                curve.hide()
      
    @QtCore.Slot(bool)
    @QtCore.Slot(int)
    def updateAlpha(self, *args):
        (alpha, auto) = self.alphaState()
        for c in self.curves:
            c.setAlpha(alpha**2, auto)
     
    def alphaState(self):
        enabled = self.ctrl.alphaGroup.isChecked()
        auto = self.ctrl.autoAlphaCheck.isChecked()
        alpha = float(self.ctrl.alphaSlider.value()) / self.ctrl.alphaSlider.maximum()
        if auto:
            alpha = 1.0  ## should be 1/number of overlapping plots
        if not enabled:
            auto = False
            alpha = 1.0
        return (alpha, auto)

    def pointMode(self):
        if self.ctrl.pointsGroup.isChecked():
            return None if self.ctrl.autoPointsCheck.isChecked() else True
        else:
            return False

    def resizeEvent(self, ev):
        if self.autoBtn is None:  ## already closed down
            return
        btnRect = self.mapRectFromItem(self.autoBtn, self.autoBtn.boundingRect())
        y = self.size().height() - btnRect.height()
        self.autoBtn.setPos(0, y)
    
    def getMenu(self):
        return self.ctrlMenu
    
    def getContextMenus(self, event):
        ## called when another item is displaying its context menu; we get to add extras to the end of the menu.
        if self.menuEnabled():
            return self.ctrlMenu
        else:
            return None
    
    def setMenuEnabled(self, enableMenu=True, enableViewBoxMenu='same'):
        """
        Enable or disable the context menu for this PlotItem.

        By default, the ViewBox's context menu will also be affected. Use 
        ``enableViewBoxMenu=None`` to leave the ViewBox unchanged.

        Parameters
        ----------
        enableMenu : bool, optional
            Whether to enable or disable the context menu, by default True.
        enableViewBoxMenu : str or bool, optional
            Whether to enable or disable the ViewBox context menu.
            If 'same', the ViewBox menu will be enabled or disabled as *enableMenu*.
            If ``True`` or ``False``, the ViewBox menu will be set accordingly.
            If ``None``, the ViewBox menu will not be changed.
            By default 'same'.
        """
        self._menuEnabled = enableMenu
        if enableViewBoxMenu is None:
            return
        if enableViewBoxMenu == 'same':
            enableViewBoxMenu = enableMenu
        self.vb.setMenuEnabled(enableViewBoxMenu)
    
    def menuEnabled(self):
        """
        Return whether the context menu is enabled.

        Returns
        -------
        bool
            ``True`` if the context menu is enabled, ``False`` otherwise.
        """
        return self._menuEnabled

    def setContextMenuActionVisible(self, name : str, visible : bool) -> None:
        """
        Change the context menu action visibility.

        Parameters
        ----------
        name : {'Transforms', 'Downsample', 'Average', 'Alpha', 'Grid', 'Points'}
            Action name.
        visible : bool
            Determines if action will be display. ``True`` action is visible, ``False``
            action is invisible.
        """
        translated_name = translate("PlotItem", name)
        for action in self.ctrlMenu.actions():
            if action.text() == translated_name:
                action.setVisible(visible)
                break
    
    def hoverEvent(self, ev):
        if ev.enter:
            self.mouseHovering = True
        if ev.exit:
            self.mouseHovering = False
            
        self.updateButtons()

    def getLabel(self, key):
        pass
        
    def _checkScaleKey(self, key):
        if key not in self.axes:
            raise KeyError(
                f"Scale '{key}' not found. Scales are: {list(self.axes.keys())}"
            )
        
    def getScale(self, key):
        return self.getAxis(key)
        
    def getAxis(self, name):
        """
        Return the specified AxisItem.

        Parameters
        ----------
        name : {'left', 'bottom', 'right', 'top'}
            The name of the axis to return.

        Returns
        -------
        AxisItem
            The :class:`~pyqtgraph.AxisItem`.

        Raises
        ------
        KeyError
            If the specified axis is not present.
        """
        self._checkScaleKey(name)
        return self.axes[name]['item']
        
    def setLabel(self, axis: str, *args, **kwargs):
        """
        Set the label for an axis.
        
        Basic HTML is allowed. See :func:`AxisItem.setLabel` for formatting options.
        
        Parameters
        ----------
        axis : {'left', 'bottom', 'right', 'top'}
            Specify which :class:`~pyqtgraph.AxisItem` to set the label for.
        *args : tuple, optional
            All extra arguments are passed to :meth:`~pyqtgraph.AxisItem.setLabel`.
        **kwargs : dict, optional
            Keyword arguments are passed to :meth:`~pyqtgraph.AxisItem.setLabel`.
        """
        self.getAxis(axis).setLabel(*args, **kwargs)
        self.showAxis(axis)
        
    def setLabels(self, **kwargs):
        """
        Set the axis labels of the plot.

        Parameters
        ----------
        **kwargs : dict, optional
            Keyword arguments are passed to  :meth:`~pyqtgraph.AxisItem.setLabel`.  The
            special keyword ``title`` can be used to set the plot title using
            :meth:`~pyqtgraph.PlotItem.setTitle`.
        """
        for k, v in kwargs.items():
            if k == 'title':
                self.setTitle(v)
            else:
                if isinstance(v, str):
                    v = (v,)
                self.setLabel(k, *v)
        
    def showLabel(self, axis, show=True):
        """
        Show or hide one of the plot's axis labels.

        Parameters
        ----------
        axis : {'left', 'bottom', 'right', 'top'}
            The axis label to show or hide.
        show : bool, optional
            Whether to show or hide the label, by default True.
        """
        self.getScale(axis).showLabel(show)

    def setTitle(self, title=None, **args):
        """
        Set the title of the plot.

        Parameters
        ----------
        title : str, optional
            The title text. If ``None``, the title will be hidden. The default is
            ``None``.
        **args : dict
            Additional keyword arguments are passed to
            :meth:`~pyqtgraph.LabelItem.setText`.
        """
        if title is None:
            self.titleLabel.setVisible(False)
            self.layout.setRowFixedHeight(0, 0)
            self.titleLabel.setMaximumHeight(0)
        else:
            self.titleLabel.setMaximumHeight(30)
            self.layout.setRowFixedHeight(0, 30)
            self.titleLabel.setVisible(True)
            self.titleLabel.setText(title, **args)

    def showAxis(self, axis: str, show: bool=True):
        """
        Show or hide an axis.

        Parameters
        ----------
        axis : {'left', 'bottom', 'right', 'top'}
            The axis to show or hide.
        show : bool, optional
            Whether to show or hide the axis, by default True.
        """
        s = self.getScale(axis)
        if show:
            s.show()
        else:
            s.hide()
            
    def hideAxis(self, axis):
        """
        Hide an axis.

        Parameters
        ----------
        axis : {'left', 'bottom', 'right', 'top'}
            The axis to hide.
        """
        self.showAxis(axis, False)
        
    def showAxes(self, selection, showValues=True, size=False):
        """ 
        Convenience method for quickly configuring axis settings.
        
        Parameters
        ----------
        selection : bool or tuple of bool 
            Determines which AxisItems will be displayed. If in tuple form, order is
            (left, top, right, bottom). A single boolean value will set all axes, so
            that ``showAxes(True)`` configures the axes to draw a frame.
        showValues : bool or tuple of bool, optional
            Determines if values will be displayed for the ticks of each axis. ``True``
            value shows values for left and bottom axis (default). ``False`` shows no
            values. If in tuple form, order is (left, top, right, bottom). ``None``
            leaves settings unchanged. If not specified, left and bottom axes will be
            drawn with values.
        size : float or tuple of float, optional
            Reserves as fixed amount of space (width for vertical axis, height for
            horizontal axis) for each axis where tick values are enabled. If only a
            single float value is given, it will be applied for both width and height.
            If ``None`` is given instead of a float value, the axis reverts to automatic
            allocation of space. If in tuple form, order is ``(width, height)``.
        """
        if selection is True: # shortcut: enable all axes, creating a frame
            selection = (True, True, True, True)
        elif selection is False: # shortcut: disable all axes
            selection = (False, False, False, False)
        if showValues is True: # shortcut: defaults arrangement with labels at left and bottom
            showValues = (True, False, False, True)
        elif showValues is False: # shortcut: disable all labels
            showValues = (False, False, False, False)
        elif showValues is None: # leave labelling untouched
            showValues = (None, None, None, None)
        if size is not False and not isinstance(size, collections.abc.Sized):
            size = (size, size) # make sure that size is either False or a full set of (width, height)

        all_axes = ('left','top','right','bottom')
        for show_axis, show_value, axis_key in zip(selection, showValues, all_axes):
            if show_axis is not None:
                if show_axis: self.showAxis(axis_key)
                else        : self.hideAxis(axis_key)

            if show_value is not None:
                ax = self.getAxis(axis_key)
                ax.setStyle(showValues=show_value)
                if size is not False: # size adjustment is requested
                    if axis_key in ('left','right'):
                        if show_value: ax.setWidth(size[0])
                        else         : ax.setWidth( None )
                    elif axis_key in ('top', 'bottom'):
                        if show_value: ax.setHeight(size[1])
                        else         : ax.setHeight( None )
  
    def hideButtons(self):
        """
        Hide auto-scale button ('A' in lower-left corner).
        """
    
        self.buttonsHidden = True
        self.updateButtons()
        
    def showButtons(self):
        """
        Show auto-scale button ('A' in lower-left corner).
        """
        self.buttonsHidden = False
        self.updateButtons()
        
    def updateButtons(self):
        try:
            if self._exportOpts is False and self.mouseHovering and not self.buttonsHidden and not all(self.vb.autoRangeEnabled()):
                self.autoBtn.show()
            else:
                self.autoBtn.hide()
        except RuntimeError:
            pass  # this can happen if the plot has been deleted.
            
    def _plotArray(self, arr, x=None, **kargs):
        if arr.ndim != 1:
            raise ValueError(f"Array must be 1D to plot (shape is {arr.shape})")
        if x is None:
            x = np.arange(arr.shape[0])
        if x.ndim != 1:
            raise ValueError(f"X array must be 1D to plot (shape is {x.shape})")
        return PlotCurveItem(arr, x=x, **kargs)

    def setExportMode(self, export: bool, opts=None):
        """
        Set whether the item will allow export via screenshots or image files.

        If export mode is enabled, then the item will allow export and the export
        options dock will be displayed. If export mode is disabled, then no export
        options will be available and the export dock will be hidden.

        Parameters
        ----------
        export : bool
            If ``True``, the item will allow export.
        opts : dict or None, optional
            A dictionary of export options. If ``None``, the default export
            options will be used. By default, ``None``.
        """
        super().setExportMode(export, opts)
        self.updateButtons()
    
    def _chooseFilenameDialog(self, handler):
        self.fileDialog = FileDialog()
        if PlotItem.lastFileDir is not None:
            self.fileDialog.setDirectory(PlotItem.lastFileDir)
        self.fileDialog.setFileMode(QtWidgets.QFileDialog.FileMode.AnyFile)
        self.fileDialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptSave)
        self.fileDialog.show()
        self.fileDialog.fileSelected.connect(handler)



import collections
import importlib.util
import re

from .. import colormap
from ..graphicsItems.GradientPresets import Gradients
from ..Qt import QtCore, QtGui, QtWidgets

__all__ = ['ColorMapMenu']


# from https://matplotlib.org/stable/gallery/color/colormap_reference.html
MATPLOTLIB_CMAPS = [
         ('Perceptually Uniform Sequential', [
            'viridis', 'plasma', 'inferno', 'magma', 'cividis']),
         ('Sequential', [
            'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
            'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
            'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']),
         ('Sequential (2)', [
            'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
            'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
            'hot', 'afmhot', 'gist_heat', 'copper']),
         ('Diverging', [
            'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
            'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic',
            'berlin', 'managua', 'vanimo']),
         ('Cyclic', ['twilight', 'twilight_shifted', 'hsv']),
         ('Qualitative', [
            'Pastel1', 'Pastel2', 'Paired', 'Accent',
            'Dark2', 'Set1', 'Set2', 'Set3',
            'tab10', 'tab20', 'tab20b', 'tab20c']),
         ('Miscellaneous', [
            'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
            'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
            'gist_rainbow', 'rainbow', 'jet', 'turbo', 'nipy_spectral',
            'gist_ncar'])
    ]


PrivateActionData = collections.namedtuple("ColorMapMenuPrivateActionData", ["name", "source"])


def buildMenuEntryWidget(cmap, text):
    lut = cmap.getLookupTable(nPts=32, alpha=True)
    qimg = QtGui.QImage(lut, len(lut), 1, QtGui.QImage.Format.Format_RGBA8888)
    pixmap = QtGui.QPixmap.fromImage(qimg)

    widget = QtWidgets.QWidget()
    layout = QtWidgets.QHBoxLayout(widget)
    layout.setContentsMargins(1,1,1,1)
    label1 = QtWidgets.QLabel()
    label1.setScaledContents(True)
    label1.setPixmap(pixmap)
    label2 = QtWidgets.QLabel(text)
    layout.addWidget(label1, 0)
    layout.addWidget(label2, 1)

    return widget

def buildMenuEntryAction(menu, name, source):
    if isinstance(source, colormap.ColorMap):
        cmap = source
    elif source == 'preset-gradient':
        cmap = preset_gradient_to_colormap(name)
    else:
        cmap = colormap.get(name, source=source)
    act = QtWidgets.QWidgetAction(menu)
    act.setData(PrivateActionData(name, source))
    act.setDefaultWidget(buildMenuEntryWidget(cmap, name))
    menu.addAction(act)

def sorted_filenames(names):
    pattern = re.compile(r'(\d+)')
    key = lambda x: [int(c) if c.isdigit() else c for c in pattern.split(x)]
    return sorted(names, key=key)

def find_mpl_leftovers():
    names = colormap.listMaps(source="matplotlib")
    # remove entries registered by colorcet
    names = [x for x in names if not x.startswith('cet_')]
    # remove the reversed colormaps
    names = [x for x in names if not x.endswith('_r')]
    # remove entries that are already categorised
    known_names = set()
    for item in MATPLOTLIB_CMAPS:
        known_names.update(item[1])
    names = [x for x in names if x not in known_names]
    return names

def buildCetSubMenu(menu, source, cet_type):
    names = colormap.listMaps(source=source)
    names = [x for x in names if x.startswith("CET")]

    if cet_type.endswith("Blind"):
        names = [x for x in names if x[4:6] == "CB"]
    else:
        names = [x for x in names if x[4] == cet_type[0] and x[5].isdigit()]

    for name in sorted_filenames(names):
        buildMenuEntryAction(menu, name, source)

def buildUserSubMenu(menu, userList):
    for item in userList:
        if isinstance(item, colormap.ColorMap):
            name, source = item.name, item
        elif isinstance(item, str):
            name, source = item, None
        elif isinstance(item, tuple):
            name, source = item
        else:
            raise ValueError("userList items must be ColorMap, str or tuple")

        buildMenuEntryAction(menu, name, source)

def preset_gradient_to_colormap(name):
    # generate the hsv two gradients using makeHslCycle
    if name == 'spectrum':
        # steps=30 for 300 degrees gives the same density as
        # the default steps=36 for 360 degrees
        cmap = colormap.makeHslCycle((0, 300/360), steps=30)
    elif name == 'cyclic':
        cmap = colormap.makeHslCycle((1, 0))
    else:
        cmap = colormap.ColorMap(*zip(*Gradients[name]["ticks"]), name=name)
    return cmap


class ColorMapMenu(QtWidgets.QMenu):
    sigColorMapTriggered = QtCore.Signal(object)

    def __init__(self, *, userList=None, showGradientSubMenu=False, showColorMapSubMenus=False):
        """
        Creates a new ColorMapMenu.

        Parameters
        ----------
        userList : list of ColorMapSpecifier, optional
            Supported values for ColorMapSpecifier are 
            ``str``, ``(str, str)``, :class:`~pyqtgraph.ColorMap`
            
            Example: ``["viridis", ("glasbey", "colorcet"), ("rainbow", "matplotlib")]``
        showGradientSubMenu : bool, default=False
            Adds legacy gradients in a submenu.
        showColorMapSubMenus : bool, default=False
            Adds bundled colormaps and external (colorcet, matplotlib) colormaps in submenus.
        """
        super().__init__()

        self.setTitle("ColorMaps")
        self.triggered.connect(self.onTriggered)

        topmenu = self
        act = topmenu.addAction('None')
        act.setData(PrivateActionData(None, None))

        if userList is not None:
            buildUserSubMenu(topmenu, userList)

        if any([showGradientSubMenu, showColorMapSubMenus]):
            topmenu.addSeparator()

        # render the submenus only if the user actually clicks on it

        if showGradientSubMenu:
            submenu = topmenu.addMenu('preset gradient')
            submenu.aboutToShow.connect(self.buildGradientSubMenu)

        if not showColorMapSubMenus:
            return

        submenu = topmenu.addMenu('local')
        submenu.aboutToShow.connect(self.buildLocalSubMenu)

        have_colorcet = importlib.util.find_spec('colorcet') is not None
        # arranged in the order listed in https://colorcet.com/
        cet_types = ["Linear", "Divergent", "Rainbow", "Cyclic", "Isoluminant", "Color Blind"]

        # the local cet files are a subset of the colorcet module.
        # expose just one of them.
        if not have_colorcet:
            submenu = topmenu.addMenu('cet (local)')
            for cet_type in cet_types:
                sub2menu = submenu.addMenu(cet_type)
                sub2menu.aboutToShow.connect(self.buildCetLocalSubMenu)
        else:
            submenu = topmenu.addMenu('cet (external)')
            for cet_type in cet_types:
                sub2menu = submenu.addMenu(cet_type)
                sub2menu.aboutToShow.connect(self.buildCetExternalSubMenu)

        if importlib.util.find_spec('matplotlib') is not None:
            submenu = topmenu.addMenu('matplotlib')
            # skip 1st entry which is "Perceptually Uniform Sequential"
            # since pyqtgraph has those already
            for category, _ in MATPLOTLIB_CMAPS[1:]:
                sub2menu = submenu.addMenu(category)
                sub2menu.aboutToShow.connect(self.buildMplCategorySubMenu)

            if find_mpl_leftovers():
                sub2menu = submenu.addMenu("Others")
                sub2menu.aboutToShow.connect(self.buildMplOthersSubMenu)

        if have_colorcet:
            submenu = topmenu.addMenu('colorcet')
            submenu.aboutToShow.connect(self.buildColorcetSubMenu)

    @QtCore.Slot(QtGui.QAction)
    def onTriggered(self, action):
        if not isinstance(data := action.data(), PrivateActionData):
            return
        cmap = self.actionDataToColorMap(data)
        self.sigColorMapTriggered.emit(cmap)

    @QtCore.Slot()
    def buildGradientSubMenu(self):
        source = 'preset-gradient'
        names = list(Gradients.keys())
        self.buildSubMenu(names, source, sort=False)

    @QtCore.Slot()
    def buildLocalSubMenu(self):
        source = None
        names = colormap.listMaps(source=source)
        names = [x for x in names if not x.startswith('CET')]
        names = [x for x in names if not x.startswith('PAL-relaxed')]
        self.buildSubMenu(names, source)

    @QtCore.Slot()
    def buildCetLocalSubMenu(self):
        # in Qt6 we could have used Qt.ConnectionType.SingleShotConnection
        menu = self.sender()
        menu.aboutToShow.disconnect()
        source = None
        cet_type = menu.title()
        buildCetSubMenu(menu, source, cet_type)

    @QtCore.Slot()
    def buildCetExternalSubMenu(self):
        # in Qt6 we could have used Qt.ConnectionType.SingleShotConnection
        menu = self.sender()
        menu.aboutToShow.disconnect()
        source = 'colorcet'
        cet_type = menu.title()
        buildCetSubMenu(menu, source, cet_type)

    @QtCore.Slot()
    def buildMplCategorySubMenu(self):
        # in Qt6 we could have used Qt.ConnectionType.SingleShotConnection
        menu = self.sender()
        menu.aboutToShow.disconnect()
        source = 'matplotlib'
        category = menu.title()
        categories = [x[0] for x in MATPLOTLIB_CMAPS]
        names = MATPLOTLIB_CMAPS[categories.index(category)][1]
        for name in names:
            try:
                buildMenuEntryAction(menu, name, source)
            except ValueError:
                # the names are not programmatically discovered,
                # so to be safe, we wrap around try except
                pass

    @QtCore.Slot()
    def buildMplOthersSubMenu(self):
        self.buildSubMenu(find_mpl_leftovers(), "matplotlib")

    @QtCore.Slot()
    def buildColorcetSubMenu(self):
        # colorcet colormaps with shorter/simpler aliases
        source = 'colorcet'
        import colorcet
        names = list(colorcet.palette_n.keys())
        self.buildSubMenu(names, source)

    def buildSubMenu(self, names, source, sort=True):
        # in Qt6 we could have used Qt.ConnectionType.SingleShotConnection
        menu = self.sender()
        menu.aboutToShow.disconnect()

        if sort:
            names = sorted_filenames(names)
        for name in names:
            buildMenuEntryAction(menu, name, source)

    @staticmethod
    def actionDataToColorMap(data):
        name, source = data
        if isinstance(source, colormap.ColorMap):
            cmap = source
        elif name is None:
            cmap = colormap.ColorMap(None, [0.0, 1.0])
        elif source == 'preset-gradient':
            cmap = preset_gradient_to_colormap(name)
            cmap.name = f"{source}:{name}"    # for GradientEditorItem
        else:
            # colormap module maintains a cache keyed by name only.
            # thus if a colormap has the same name in two different sources,
            # we will end up getting whatever was already cached.
            cmap = colormap.get(name, source=source)
        return cmap



"""
This module exists to smooth out some of the differences between Qt versions.

* Automatically import Qt lib depending on availability
* Allow you to import QtCore/QtGui from pyqtgraph.Qt without specifying which Qt wrapper
  you want to use.
"""
import contextlib
import os
import platform
import re
import subprocess
import sys
import time
import warnings
from importlib import resources

PYSIDE = 'PySide'
PYSIDE2 = 'PySide2'
PYSIDE6 = 'PySide6'
PYQT4 = 'PyQt4'
PYQT5 = 'PyQt5'
PYQT6 = 'PyQt6'

QT_LIB = os.getenv('PYQTGRAPH_QT_LIB')

if QT_LIB is not None:
    try:
        __import__(QT_LIB)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"Environment variable PYQTGRAPH_QT_LIB is set to '{os.getenv('PYQTGRAPH_QT_LIB')}', but no module with this name was found.")

## Automatically determine which Qt package to use (unless specified by
## environment variable).
## This is done by first checking to see whether one of the libraries
## is already imported. If not, then attempt to import in the order
## specified in libOrder.
if QT_LIB is None:
    libOrder = [PYQT6, PYSIDE6, PYQT5, PYSIDE2]

    for lib in libOrder:
        if lib in sys.modules:
            QT_LIB = lib
            break

if QT_LIB is None:
    for lib in libOrder:
        qt = lib + '.QtCore'
        try:
            __import__(qt)
            QT_LIB = lib
            break
        except ImportError:
            pass

if QT_LIB is None:
    raise ImportError("PyQtGraph requires one of PyQt5, PyQt6, PySide2 or PySide6; none of these packages could be imported.")


class FailedImport(object):
    """Used to defer ImportErrors until we are sure the module is needed.
    """
    def __init__(self, err):
        self.err = err
        
    def __getattr__(self, attr):
        raise self.err


# Make a loadUiType function like PyQt has

# Credit:
# http://stackoverflow.com/questions/4442286/python-code-genration-with-pyside-uic/14195313#14195313

class _StringIO(object):
    """Alternative to built-in StringIO needed to circumvent unicode/ascii issues"""
    def __init__(self):
        self.data = []
    
    def write(self, data):
        self.data.append(data)
        
    def getvalue(self):
        return ''.join(map(str, self.data)).encode('utf8')

    
def _loadUiType(uiFile):
    """
    PySide lacks a "loadUiType" command like PyQt4's, so we have to convert
    the ui file to py code in-memory first and then execute it in a
    special frame to retrieve the form_class.

    from stackoverflow: http://stackoverflow.com/a/14195313/3781327

    seems like this might also be a legitimate solution, but I'm not sure
    how to make PyQt4 and pyside look the same...
        http://stackoverflow.com/a/8717832
    """

    pyside2uic = None
    if QT_LIB == PYSIDE2:
        try:
            import pyside2uic
        except ImportError:
            # later versions of pyside2 have dropped pyside2uic; use the uic binary instead.
            pyside2uic = None

        if pyside2uic is None:
            pyside2version = tuple(map(int, PySide2.__version__.split(".")))
            if (5, 14) <= pyside2version < (5, 14, 2, 2):
                warnings.warn('For UI compilation, it is recommended to upgrade to PySide >= 5.15', RuntimeWarning, stacklevel=2)

    # get class names from ui file
    import xml.etree.ElementTree as xml
    parsed = xml.parse(uiFile)
    widget_class = parsed.find('widget').get('class')
    form_class = parsed.find('class').text

    # convert ui file to python code
    if pyside2uic is None:
        uic_executable = QT_LIB.lower() + '-uic'
        uipy = subprocess.check_output([uic_executable, uiFile])
    else:
        o = _StringIO()
        with open(uiFile, 'r') as f:
            pyside2uic.compileUi(f, o, indent=0)
        uipy = o.getvalue()

    # execute python code
    pyc = compile(uipy, '<string>', 'exec')
    frame = {}
    exec(pyc, frame)

    # fetch the base_class and form class based on their type in the xml from designer
    form_class = frame['Ui_%s'%form_class]
    base_class = eval('QtWidgets.%s'%widget_class)

    return form_class, base_class


# For historical reasons, pyqtgraph maintains a Qt4-ish interface back when
# there wasn't a QtWidgets module. This _was_ done by monkey-patching all of
# QtWidgets into the QtGui module. This monkey-patching modifies QtGui at a
# global level.
# To avoid this, we now maintain a local "mirror" of QtCore, QtGui and QtWidgets.
# Thus, when monkey-patching happens later on in this file, they will only affect
# the local modules and not the global modules.
def _copy_attrs(src, dst):
    for o in dir(src):
        if not hasattr(dst, o):
            setattr(dst, o, getattr(src, o))

from . import QtCore, QtGui, QtWidgets, compat

if QT_LIB == PYQT5:
    # We're using PyQt5 which has a different structure so we're going to use a shim to
    # recreate the Qt4 structure for Qt5
    import PyQt5.QtCore
    import PyQt5.QtGui
    import PyQt5.QtWidgets
    _copy_attrs(PyQt5.QtCore, QtCore)
    _copy_attrs(PyQt5.QtGui, QtGui)
    _copy_attrs(PyQt5.QtWidgets, QtWidgets)

    try:
        from PyQt5 import sip
    except ImportError:
        # some Linux distros package it this way (e.g. Ubuntu)
        import sip
    from PyQt5 import uic

    try:
        from PyQt5 import QtSvg
    except ImportError as err:
        QtSvg = FailedImport(err)
    try:
        from PyQt5 import QtTest
    except ImportError as err:
        QtTest = FailedImport(err)

    VERSION_INFO = 'PyQt5 ' + QtCore.PYQT_VERSION_STR + ' Qt ' + QtCore.QT_VERSION_STR

elif QT_LIB == PYQT6:
    import PyQt6.QtCore
    import PyQt6.QtGui
    import PyQt6.QtWidgets
    _copy_attrs(PyQt6.QtCore, QtCore)
    _copy_attrs(PyQt6.QtGui, QtGui)
    _copy_attrs(PyQt6.QtWidgets, QtWidgets)

    from PyQt6 import sip, uic

    try:
        from PyQt6 import QtSvg
    except ImportError as err:
        QtSvg = FailedImport(err)
    try:
        from PyQt6 import QtOpenGLWidgets
    except ImportError as err:
        QtOpenGLWidgets = FailedImport(err)
    try:
        from PyQt6 import QtTest
    except ImportError as err:
        QtTest = FailedImport(err)

    VERSION_INFO = 'PyQt6 ' + QtCore.PYQT_VERSION_STR + ' Qt ' + QtCore.QT_VERSION_STR

elif QT_LIB == PYSIDE2:
    import PySide2.QtCore
    import PySide2.QtGui
    import PySide2.QtWidgets
    _copy_attrs(PySide2.QtCore, QtCore)
    _copy_attrs(PySide2.QtGui, QtGui)
    _copy_attrs(PySide2.QtWidgets, QtWidgets)
    
    try:
        from PySide2 import QtSvg
    except ImportError as err:
        QtSvg = FailedImport(err)
    try:
        from PySide2 import QtTest
    except ImportError as err:
        QtTest = FailedImport(err)

    import PySide2
    import shiboken2 as shiboken
    VERSION_INFO = 'PySide2 ' + PySide2.__version__ + ' Qt ' + QtCore.__version__
elif QT_LIB == PYSIDE6:
    import PySide6.QtCore
    import PySide6.QtGui
    import PySide6.QtWidgets
    _copy_attrs(PySide6.QtCore, QtCore)
    _copy_attrs(PySide6.QtGui, QtGui)
    _copy_attrs(PySide6.QtWidgets, QtWidgets)

    try:
        from PySide6 import QtSvg
    except ImportError as err:
        QtSvg = FailedImport(err)
    try:
        from PySide6 import QtOpenGLWidgets
    except ImportError as err:
        QtOpenGLWidgets = FailedImport(err)
    try:
        from PySide6 import QtTest
    except ImportError as err:
        QtTest = FailedImport(err)

    import PySide6
    import shiboken6 as shiboken
    VERSION_INFO = 'PySide6 ' + PySide6.__version__ + ' Qt ' + QtCore.__version__

else:
    raise ValueError("Invalid Qt lib '%s'" % QT_LIB)



if QT_LIB in [PYQT6, PYSIDE6]:
    # We're using Qt6 which has a different structure so we're going to use a shim to
    # recreate the Qt5 structure

    if not isinstance(QtOpenGLWidgets, FailedImport):
        QtWidgets.QOpenGLWidget = QtOpenGLWidgets.QOpenGLWidget

    # PySide6 incorrectly placed QFileSystemModel inside QtWidgets
    if QT_LIB == PYSIDE6 and hasattr(QtWidgets, 'QFileSystemModel'):
        module = getattr(QtWidgets, "QFileSystemModel")
        setattr(QtGui, "QFileSystemModel", module)

else:
    # Shim Qt5 namespace to match Qt6
    module_whitelist = [
        "QAction",
        "QActionGroup",
        "QFileSystemModel",
        "QShortcut",
        "QUndoCommand",
        "QUndoGroup",
        "QUndoStack",
    ]
    for module in module_whitelist:
        attr = getattr(QtWidgets, module)
        setattr(QtGui, module, attr)


# Common to PySide2 and PySide6
if QT_LIB in [PYSIDE2, PYSIDE6]:
    QtVersion = QtCore.__version__
    loadUiType = _loadUiType
    isQObjectAlive = shiboken.isValid

    # PySide does not implement qWait
    if not isinstance(QtTest, FailedImport):
        if not hasattr(QtTest.QTest, 'qWait'):
            @staticmethod
            def qWait(msec):
                start = time.time()
                QtWidgets.QApplication.processEvents()
                while time.time() < start + msec * 0.001:
                    QtWidgets.QApplication.processEvents()
            QtTest.QTest.qWait = qWait

    compat.wrapinstance = shiboken.wrapInstance
    compat.unwrapinstance = lambda x : shiboken.getCppPointer(x)[0]
    compat.voidptr = shiboken.VoidPtr

# Common to PyQt5 and PyQt6
if QT_LIB in [PYQT5, PYQT6]:
    QtVersion = QtCore.QT_VERSION_STR

    # PyQt, starting in v5.5, calls qAbort when an exception is raised inside
    # a slot. To maintain backward compatibility (and sanity for interactive
    # users), we install a global exception hook to override this behavior.
    if sys.excepthook == sys.__excepthook__:
        sys_excepthook = sys.excepthook
        def pyqt_qabort_override(*args, **kwds):
            return sys_excepthook(*args, **kwds)
        sys.excepthook = pyqt_qabort_override
    
    def isQObjectAlive(obj):
        return not sip.isdeleted(obj)
    
    loadUiType = uic.loadUiType

    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot

    compat.wrapinstance = sip.wrapinstance
    compat.unwrapinstance = sip.unwrapinstance
    compat.voidptr = sip.voidptr

from . import internals

# Alert user if using Qt < 5.15, but do not raise exception
versionReq = [5, 15]
m = re.match(r'(\d+)\.(\d+).*', QtVersion)
if m is not None and list(map(int, m.groups())) < versionReq:
    warnings.warn(
        f"PyQtGraph supports Qt version >= {versionReq[0]}.{versionReq[1]},"
        f" but {QtVersion} detected.",
        RuntimeWarning,
        stacklevel=2
    )

App = QtWidgets.QApplication
# subclassing QApplication causes segfaults on PySide{2, 6} / Python 3.8.7+

QAPP = None
_pgAppInitialized = False
def mkQApp(name=None):
    """
    Creates new QApplication or returns current instance if existing.
    
    ============== ========================================================
    **Arguments:**
    name           (str) Application name, passed to Qt
    ============== ========================================================
    """
    global QAPP
    global _pgAppInitialized

    QAPP = QtWidgets.QApplication.instance()
    if QAPP is None:
        # We do not have an already instantiated QApplication
        # let's add some sane defaults

        # hidpi handling
        qtVersionCompare = tuple(map(int, QtVersion.split(".")))
        if qtVersionCompare > (6, 0):
            # Qt6 seems to support hidpi without needing to do anything so continue
            pass
        elif qtVersionCompare > (5, 14):
            os.environ["QT_ENABLE_HIGHDPI_SCALING"] = "1"
            QtWidgets.QApplication.setHighDpiScaleFactorRoundingPolicy(
                QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
            )
        else:  # qt 5.12 and 5.13
            QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
            QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

        QAPP = QtWidgets.QApplication(sys.argv or ["pyqtgraph"])
        if QtVersion.startswith("6"):
            # issues with dark mode + windows + qt5
            QAPP.setStyle("fusion")

        # set the application icon
        # python 3.9 won't take "pyqtgraph.icons.peegee" directly
        traverse_path = resources.files("pyqtgraph.icons")  
        peegee_traverse_path = traverse_path.joinpath("peegee")

        # as_file requires I feed in a file from the directory...
        with resources.as_file(
            peegee_traverse_path.joinpath("peegee.svg")
        ) as path:
            # need the parent directory, not the filepath
            icon_path = path.parent

        applicationIcon = QtGui.QIcon()
        applicationIcon.addFile(
            os.fsdecode(icon_path / "peegee.svg"),
        )
        for sz in [128, 256, 512]:
            pathname = os.fsdecode(icon_path / f"peegee_{sz}px.png")
            applicationIcon.addFile(pathname, QtCore.QSize(sz, sz))

        # handles the icon showing up on the windows taskbar
        if platform.system() == 'Windows':
            import ctypes
            my_app_id = "pyqtgraph.Qt.mkQApp"
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(my_app_id)
        QAPP.setWindowIcon(applicationIcon)

    if not _pgAppInitialized:
        _pgAppInitialized = True

        # determine if dark mode
        try:
            # this only works in Qt 6.5+
            darkMode = QAPP.styleHints().colorScheme() == QtCore.Qt.ColorScheme.Dark
            QAPP.styleHints().colorSchemeChanged.connect(_onColorSchemeChange)
        except AttributeError:
            palette = QAPP.palette()
            windowTextLightness = palette.color(QtGui.QPalette.ColorRole.WindowText).lightness()
            windowLightness = palette.color(QtGui.QPalette.ColorRole.Window).lightness()
            darkMode = windowTextLightness > windowLightness
            QAPP.paletteChanged.connect(_onPaletteChange)
        QAPP.setProperty("darkMode", darkMode)

    if name is not None:
        QAPP.setApplicationName(name)
    return QAPP


def _onPaletteChange(palette):
    # Attempt to keep darkMode attribute up to date
    # QEvent.Type.PaletteChanged/ApplicationPaletteChanged will be emitted after
    # paletteChanged.emit()!
    # Using API deprecated in Qt 6.0
    app = mkQApp()
    windowTextLightness = palette.color(QtGui.QPalette.ColorRole.WindowText).lightness()
    windowLightness = palette.color(QtGui.QPalette.ColorRole.Window).lightness()
    darkMode = windowTextLightness > windowLightness
    app.setProperty('darkMode', darkMode)


def _onColorSchemeChange(colorScheme):
    # Attempt to keep darkMode attribute up to date
    # QEvent.Type.PaletteChanged/ApplicationPaletteChanged will be emitted before
    # QStyleHint().colorSchemeChanged.emit()!
    # Uses Qt 6.5+ API
    app = mkQApp()
    darkMode = colorScheme == QtCore.Qt.ColorScheme.Dark
    app.setProperty('darkMode', darkMode)


# exec() is used within _loadUiType, so we define as exec_() here and rename in pg namespace
def exec_():
    app = mkQApp()
    return app.exec() if hasattr(app, 'exec') else app.exec_()



from ..Qt import QtGui
from ..Qt import OpenGLConstants as GLC
from ..Qt import OpenGLHelpers


def print_version(funcs):
    glGetString = funcs.glGetString
    print('VENDOR:', glGetString(GLC.GL_VENDOR))
    print('RENDERER:', glGetString(GLC.GL_RENDERER))
    print('VERSION:', glGetString(GLC.GL_VERSION))
    print('GLSL_VERSION:', glGetString(GLC.GL_SHADING_LANGUAGE_VERSION))


def print_extensions(ctx):
    extensions = sorted([ext.data().decode() for ext in ctx.extensions()])
    print("Extensions:")
    for ext in extensions:
        print(f"   {ext}")


app = QtGui.QGuiApplication([])
surf = QtGui.QOffscreenSurface()
surf.create()
ctx = QtGui.QOpenGLContext()
ctx.create()
ctx.makeCurrent(surf)

print("openGLModuleType:", QtGui.QOpenGLContext.openGLModuleType())
print('isOpenGLES:', ctx.isOpenGLES())
glfn = OpenGLHelpers.getFunctions(ctx)
print_version(glfn)
print_extensions(ctx)



"""
Use GLImageItem to display image data on rectangular planes.

In this example, the image data is sampled from a volume and the image planes 
placed as if they slice through the volume.
"""
import sys

import numpy as np

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui
import pyqtgraph.opengl as gl

if 'darwin' in sys.platform:
    fmt = QtGui.QSurfaceFormat()
    fmt.setRenderableType(fmt.RenderableType.OpenGL)
    fmt.setProfile(fmt.OpenGLContextProfile.CoreProfile)
    fmt.setVersion(4, 1)
    QtGui.QSurfaceFormat.setDefaultFormat(fmt)

app = pg.mkQApp("GLImageItem Example")
w = gl.GLViewWidget()
w.show()
w.setWindowTitle('pyqtgraph example: GLImageItem')
w.setCameraPosition(distance=200)

## create volume data set to slice three images from
shape = (100,100,70)
data = pg.gaussianFilter(np.random.normal(size=shape), (4,4,4))
data += pg.gaussianFilter(np.random.normal(size=shape), (15,15,15))*15

## slice out three planes, convert to RGBA for OpenGL texture
levels = (-0.08, 0.08)
tex1 = pg.makeRGBA(data[shape[0]//2], levels=levels)[0]       # yz plane
tex2 = pg.makeRGBA(data[:,shape[1]//2], levels=levels)[0]     # xz plane
tex3 = pg.makeRGBA(data[:,:,shape[2]//2], levels=levels)[0]   # xy plane
#tex1[:,:,3] = 128
#tex2[:,:,3] = 128
#tex3[:,:,3] = 128

## Create three image items from textures, add to view
v1 = gl.GLImageItem(tex1)
v1.translate(-shape[1]/2, -shape[2]/2, 0)
v1.rotate(90, 0,0,1)
v1.rotate(-90, 0,1,0)
w.addItem(v1)
v2 = gl.GLImageItem(tex2)
v2.translate(-shape[0]/2, -shape[2]/2, 0)
v2.rotate(-90, 1,0,0)
w.addItem(v2)
v3 = gl.GLImageItem(tex3)
v3.translate(-shape[0]/2, -shape[1]/2, 0)
w.addItem(v3)

ax = gl.GLAxisItem()
w.addItem(ax)

if __name__ == '__main__':
    pg.exec()



"""
Demonstrates GLVolumeItem for displaying volumetric data.
"""
import sys

import numpy as np

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui
import pyqtgraph.opengl as gl
from pyqtgraph import functions as fn

if 'darwin' in sys.platform:
    fmt = QtGui.QSurfaceFormat()
    fmt.setRenderableType(fmt.RenderableType.OpenGL)
    fmt.setProfile(fmt.OpenGLContextProfile.CoreProfile)
    fmt.setVersion(4, 1)
    QtGui.QSurfaceFormat.setDefaultFormat(fmt)

app = pg.mkQApp("GLVolumeItem Example")
w = gl.GLViewWidget()
w.show()
w.setWindowTitle('pyqtgraph example: GLVolumeItem')
w.setCameraPosition(distance=200)

g = gl.GLGridItem()
g.scale(10, 10, 1)
w.addItem(g)

## Hydrogen electron probability density
def psi(i, j, k, offset=(50,50,100)):
    x = i-offset[0]
    y = j-offset[1]
    z = k-offset[2]
    th = np.arctan2(z, np.hypot(x, y))
    r = np.sqrt(x**2 + y**2 + z **2)
    a0 = 2
    return (
        (1.0 / 81.0)
        * 1.0 / (6.0 * np.pi) ** 0.5
        * (1.0 / a0) ** (3 / 2)
        * (r / a0) ** 2
        * np.exp(-r / (3 * a0))
        * (3 * np.cos(th) ** 2 - 1)
    )


data = np.fromfunction(psi, (100,100,200))
with np.errstate(divide = 'ignore'):
    positive = np.log(fn.clip_array(data, 0, data.max())**2)
    negative = np.log(fn.clip_array(-data, 0, -data.min())**2)

d2 = np.empty(data.shape + (4,), dtype=np.ubyte)

# Original Code
# d2[..., 0] = positive * (255./positive.max())
# d2[..., 1] = negative * (255./negative.max())

# Reformulated Code
# Both positive.max() and negative.max() are negative-valued.
# Thus the next 2 lines are _not_ bounded to [0, 255]
positive = positive * (255./positive.max())
negative = negative * (255./negative.max())
# When casting to ubyte, the original code relied on +Inf to be
# converted to 0. On arm64, it gets converted to 255.
# Thus the next 2 lines change +Inf explicitly to 0 instead.
positive[np.isinf(positive)] = 0
negative[np.isinf(negative)] = 0
# When casting to ubyte, the original code relied on the conversion
# to do modulo 256. The next 2 lines do it explicitly instead as
# documentation.
d2[..., 0] = positive.astype(int) % 256
d2[..., 1] = negative.astype(int) % 256

d2[..., 2] = d2[...,1]
d2[..., 3] = d2[..., 0]*0.3 + d2[..., 1]*0.3
d2[..., 3] = (d2[..., 3].astype(float) / 255.) **2 * 255

d2[:, 0, 0] = [255,0,0,100]
d2[0, :, 0] = [0,255,0,100]
d2[0, 0, :] = [0,0,255,100]

v = gl.GLVolumeItem(d2)
v.translate(-50,-50,-100)
w.addItem(v)

ax = gl.GLAxisItem()
w.addItem(ax)

if __name__ == '__main__':
    pg.exec()



import numpy

from .Qt import QtGui
from . import functions
from .util.cupy_helper import getCupy
from .util.numba_helper import getNumbaFunctions


def _apply_lut_for_uint(xp, image, lut):
    # Note: compared to makeARGB(), we have already clipped the data to range

    # if lut is 1d, then lut[image] is fastest
    # if lut is 2d, then lut.take(image, axis=0) is faster than lut[image]
    lut = _convert_2dlut_to_1dlut(xp, lut)

    if xp == numpy and (fn_numba := getNumbaFunctions()) is not None:
        # numba "take" supports only the 1st 2 arguments of np.take,
        # therefore we have to convert the lut to 1d.
        # "take" will output a c contiguous array regardless of its input.
        image = fn_numba.numba_take(lut, image)
    else:
        # advanced indexing is memory order aware.
        # its output can be either C or F contiguous.
        image = lut[image]

    if image.dtype == xp.uint32:
        # "view" requires c contiguous for numpy < 1.23
        image = xp.ascontiguousarray(image)
        image = image[..., xp.newaxis].view(xp.uint8)

    return image


def _convert_lut_to_rgba(xp, lut):
    # converts:
    #   - None to (256, 4)
    #   - uint8 (N,) to uint8 (N, 4)
    #   - uint8 (N, 1) to uint8 (N, 4)
    #   - uint8 (N, 3) to uint8 (N, 4)

    if not (
        lut is None
        or lut.ndim == 1
        or (
            lut.ndim == 2
            and lut.shape[1] in (1, 3, 4)
        )
    ):
        raise ValueError("unsupported lut shape")

    N = lut.shape[0] if lut is not None else 256

    if lut is None:
        lut = xp.arange(N, dtype=xp.uint8)

    # convert (N,) to (N, 1)
    if lut.ndim == 1:
        lut = lut[:, xp.newaxis]

    if lut.shape[1] == 4:
        return lut

    out = xp.full((N, 4), 255, dtype=xp.uint8)
    out[:, 0:3] = lut
    return out


def _convert_2dlut_to_1dlut(xp, lut):
    # converts:
    #   - uint8 (N, 1) to uint8 (N,)
    #   - uint8 (N, 3) or (N, 4) to uint32 (N,)
    # this allows faster lookup as 1d lookup is faster

    if lut.ndim == 1:
        return lut

    if lut.shape[1] == 3:  # rgb
        # convert rgb lut to rgba so that it is 32-bits
        lut = xp.column_stack([lut, xp.full(lut.shape[0], 255, dtype=xp.uint8)])
    if lut.shape[1] == 4:  # rgba
        lut = lut.view(xp.uint32)
    lut = lut.ravel()

    return lut


def _rescale_and_lookup_float(xp, image, levels, lut, *, forceApplyLut):
    # It is usually more performant to _not_ apply the lut and
    # instead use it as an Indexed8 ColorTable. This is only
    # applicable if the lut has <= 256 entries.

    if forceApplyLut and lut is None:
        raise ValueError("forceApplyLut True but lut not provided")

    # Decide on maximum scaled value
    if lut is not None:
        num_colors = lut.shape[0]
        max_scale_value = num_colors
    else:
        num_colors = 256
        max_scale_value = 255.0
    dtype = xp.min_scalar_type(num_colors - 1)

    # note: "dtype == uint16" ==> lut provided ==> mono-channel image
    #       i.e. multi-channel image ==> lut is None ==> dtype == uint8
    #
    #       the library defaults to using 256-entry luts, so
    #       "dtype == uint8" is the common case

    apply_lut = forceApplyLut or dtype == xp.uint16

    minVal, maxVal = levels
    rng = maxVal - minVal
    rng = 1 if rng == 0 else rng
    offset = minVal
    scale = max_scale_value / rng

    if xp == numpy and (fn_numba := getNumbaFunctions()) is not None:
        if apply_lut:
            # this path does rescale and apply lut in one step
            lut = _convert_2dlut_to_1dlut(xp, lut)
            image = fn_numba.rescale_and_lookup(image, scale, offset, lut)
            lut = None
            if image.dtype == xp.uint32:
                # "view" requires c contiguous for numpy < 1.23
                image = xp.ascontiguousarray(image)
                image = image[..., xp.newaxis].view(xp.uint8)
        else:
            image = fn_numba.rescale_and_clip(image, scale, offset, 0, num_colors - 1)
    else:
        image = functions.rescaleData(
            image, scale, offset, dtype=dtype, clip=(0, num_colors - 1)
        )
        if apply_lut:
            image = _apply_lut_for_uint(xp, image, lut)
            lut = None

    # image is now of type uint8
    return image, lut


def _combine_levels_and_lut(xp, image, levels, lut):
    if (
        image.dtype == xp.uint16
        and levels is None
        and image.ndim == 3
        and image.shape[2] == 3
    ):
        # uint16 rgb can't be directly displayed, so make it
        # pass through effective lut processing
        levels = [0, 65535]

    if levels is None and lut is None:
        # nothing to combine
        return image, lut

    # distinguish between lut for levels and colors
    levels_lut = None
    colors_lut = lut

    eflsize = 2 ** (image.itemsize * 8)
    if levels is None:
        info = xp.iinfo(image.dtype)
        minlev, maxlev = info.min, info.max
    else:
        minlev, maxlev = levels
    levdiff = maxlev - minlev
    levdiff = 1 if levdiff == 0 else levdiff  # don't allow division by 0
    offset = minlev

    if colors_lut is None:
        scale = 255.0 / levdiff
        if image.dtype == xp.ubyte and image.ndim == 2:
            # uint8 mono image
            ind = xp.arange(eflsize)
            levels_lut = functions.rescaleData(ind, scale, offset, dtype=xp.ubyte)
            # image data is not scaled. instead, levels_lut is used
            # as (grayscale) Indexed8 ColorTable to get the same effect.
            # due to the small size of the input to rescaleData(), we
            # do not bother caching the result
            return image, levels_lut
        else:
            # uint16 mono, uint8 rgb, uint16 rgb
            # rescale image data by computation instead of by memory lookup
            if xp == numpy and (fn_numba := getNumbaFunctions()) is not None:
                image = fn_numba.rescale_and_clip(image, scale, offset, 0, 255)
            else:
                image = functions.rescaleData(image, scale, offset, dtype=xp.ubyte)
            return image, colors_lut
    else:
        num_colors = colors_lut.shape[0]
        scale = num_colors / levdiff
        lutdtype = xp.min_scalar_type(num_colors - 1)

        if image.dtype == xp.ubyte or lutdtype != xp.ubyte:
            # combine if either:
            #   1) uint8 mono image
            #   2) colors_lut has more entries than will fit within 8-bits
            ind = xp.arange(eflsize)
            levels_lut = functions.rescaleData(
                ind, scale, offset, dtype=lutdtype, clip=(0, num_colors - 1),
            )
            efflut = colors_lut[levels_lut]

            # apply the effective lut early for the following types:
            if image.dtype == xp.uint16 and image.ndim == 2:
                image = _apply_lut_for_uint(xp, image, efflut)
                efflut = None
            return image, efflut
        else:
            # uint16 image with colors_lut <= 256 entries
            # don't combine, we will use QImage ColorTable
            if xp == numpy and (fn_numba := getNumbaFunctions()) is not None:
                image = fn_numba.rescale_and_clip(image, scale, offset, 0, num_colors - 1)
            else:
                image = functions.rescaleData(
                    image, scale, offset, dtype=lutdtype, clip=(0, num_colors - 1),
                )
            return image, colors_lut


def try_make_qimage(image, *, levels, lut, transparentLocations=None):
    """
    Internal function to make an QImage from an ndarray without going
    through the full generality of makeARGB().
    Only certain combinations of input arguments are supported.
    """

    # this function assumes that image has no nans.
    # checking for nans is an expensive operation; it is expected that
    # the caller would want to cache the result rather than have this
    # function check for nans unconditionally.

    cp = getCupy()
    xp = cp.get_array_module(image) if cp else numpy

    # float images always need levels
    if image.dtype.kind == "f" and levels is None:
        return None

    if levels is not None:
        levels = xp.asarray(levels)

        # can't handle multi-channel levels
        if levels.ndim != 1:
            return None

        # if levels is provided, multi-channel images must be 3 channels only.
        # (because it doesn't make sense to scale a 4th alpha channel.)
        if image.ndim == 3 and image.shape[2] != 3:
            return None

    if lut is not None and lut.dtype != xp.uint8:
        raise ValueError("lut dtype must be uint8")

    alpha_channel_required = (
        (   # image itself has alpha channel
            image.ndim == 3
            and image.shape[2] == 4
        )
        or
        (    # lut has alpha channel
            lut is not None
            and lut.ndim == 2
            and lut.shape[1] == 4
        )
    )

    if image.dtype.kind == "f":
        if image.ndim == 2:
            # mono float images
            if transparentLocations is None:
                image, lut = _rescale_and_lookup_float(
                    xp, image, levels, lut, forceApplyLut=False
                )
                levels = None
                # on return, we will have an uint8 image.
                # lut if not None will have <= 256 entries
            else:
                # this path creates an alpha channel
                lut = _convert_lut_to_rgba(xp, lut)
                alpha_channel_required = True

                image, lut = _rescale_and_lookup_float(
                    xp, image, levels, lut, forceApplyLut=True
                )
                levels = None
                assert lut is None
                image[..., 3][transparentLocations] = 0
        else:
            # RGB float images
            # lut can only be None for RGB images
            image, lut = _rescale_and_lookup_float(
                xp, image, levels, lut, forceApplyLut=False
            )
            levels = None

            if transparentLocations is not None:
                alpha_channel_required = True
                mask = xp.full(image.shape[:2], 255, dtype=xp.uint8)
                mask[transparentLocations] = 0
                image = xp.dstack((image, mask))

    # if the image data is a small int, then we can combine levels + lut
    # into a single lut for better performance
    elif image.dtype in (xp.ubyte, xp.uint16):
        image, lut = _combine_levels_and_lut(xp, image, levels, lut)
        levels = None

    ubyte_nolvl = image.dtype == xp.ubyte and levels is None
    is_passthru8 = ubyte_nolvl and lut is None
    is_indexed8 = (
        ubyte_nolvl and image.ndim == 2 and lut is not None and lut.shape[0] <= 256
    )
    is_passthru16 = image.dtype == xp.uint16 and levels is None and lut is None
    can_grayscale16 = (
        is_passthru16
        and image.ndim == 2
        and hasattr(QtGui.QImage.Format, "Format_Grayscale16")
    )
    is_rgba64 = is_passthru16 and image.ndim == 3 and image.shape[2] == 4

    # bypass makeARGB for supported combinations
    supported = is_passthru8 or is_indexed8 or can_grayscale16 or is_rgba64
    if not supported:
        return None

    if xp == cp:
        image = image.get()

    # worthwhile supporting non-contiguous arrays
    image = numpy.ascontiguousarray(image)

    fmt = None
    ctbl = None
    if is_passthru8:
        # both levels and lut are None
        # these images are suitable for display directly
        if image.ndim == 2:
            fmt = QtGui.QImage.Format.Format_Grayscale8
        elif image.shape[2] == 3:
            fmt = QtGui.QImage.Format.Format_RGB888
        elif image.shape[2] == 4:
            if alpha_channel_required:
                fmt = QtGui.QImage.Format.Format_RGBA8888
            else:
                fmt = QtGui.QImage.Format.Format_RGBX8888
    elif is_indexed8:
        # levels and/or lut --> lut-only
        fmt = QtGui.QImage.Format.Format_Indexed8
        if lut.ndim == 1 or lut.shape[1] == 1:
            ctbl = [QtGui.qRgb(x, x, x) for x in lut.ravel().tolist()]
        elif lut.shape[1] == 3:
            ctbl = [QtGui.qRgb(*rgb) for rgb in lut.tolist()]
        elif lut.shape[1] == 4:
            ctbl = [QtGui.qRgba(*rgba) for rgba in lut.tolist()]
    elif can_grayscale16:
        # single channel uint16
        # both levels and lut are None
        fmt = QtGui.QImage.Format.Format_Grayscale16
    elif is_rgba64:
        # uint16 rgba
        # both levels and lut are None
        fmt = QtGui.QImage.Format.Format_RGBA64  # endian-independent
    if fmt is None:
        raise ValueError("unsupported image type")
    qimage = functions.ndarray_to_qimage(image, fmt)
    if ctbl is not None:
        qimage.setColorTable(ctbl)
    return qimage
